"""
Looking up addresses for function names / symbols, and
vice-versa.

Uses IDA when available if there isn't sufficient symbol
information available.
"""

from __future__ import annotations

import gdb

import pwndbg.lib.cache


@pwndbg.lib.cache.cache_until("stop", "start")
def selected_frame_source_absolute_filename():
    """
    Retrieve the symbol table’s source absolute file name from the selected frame.

    In case of missing symbol table or frame information, None is returned.
    """
    try:
        frame = gdb.selected_frame()
    except gdb.error:
        return None

    if not frame:
        return None

    sal = frame.find_sal()
    if not sal:
        return None

    symtab = sal.symtab
    if not symtab:
        return None

    return symtab.fullname()



"""
IPython interactive helpers for pwndbg.

Provides convenient shortcuts for common debugging operations.
"""

from __future__ import annotations

from typing import Any

import pwndbg.aglib.memory
import pwndbg.aglib.vmmap
import pwndbg.hexdump
import pwndbg.search
from pwndbg.lib.tips import color_tip

mr = pwndbg.aglib.memory.read
mw = pwndbg.aglib.memory.write
regs = pwndbg.aglib.regs
rr = pwndbg.aglib.regs.read_reg
rw = pwndbg.aglib.regs.write_reg


def hd(addr: int, count: int = 0x40) -> None:
    """HexDump - Hexdump memory at address

    Args:
        addr: Address to dump
        count: Number of bytes (default: 0x40)

    Example:
        hd(0x400000)
        hd(0x400000, 0x100)
    """
    data = pwndbg.aglib.memory.read(addr, count)
    for line in pwndbg.hexdump.hexdump(bytes(data), address=addr, count=count):
        print(line)


def ms(
    pattern: bytes,
    start: int | None = None,
    end: int | None = None,
    limit: int = 100,
    show: bool = True,
) -> list[int]:
    """Search Memory for byte pattern

    Args:
        pattern: Bytes to search for
        start: Start address (optional)
        end: End address (optional)
        limit: Max results (default: 100)
        show: Print results

    Returns:
        List of addresses where pattern found

    Example:
        ms(b"ELF")
        ms(b"\\x90\\x90", limit=10)
    """
    results = list(pwndbg.search.search(searchfor=pattern, start=start, end=end, limit=limit))

    if show:
        for addr in results:
            print(f"{addr:#x}")

    return results


def vm(show: bool = True) -> tuple[Any, ...]:
    """Virtual memory Map - Show memory mappings.

    Args:
        show: Print formatted output

    Returns:
        Tuple of memory pages

    Example:
        vm()           # Print vmmap
        pages = vm(show=False)  # Get raw data
    """
    pages = pwndbg.aglib.vmmap.get()

    if show:
        print("Address Range          Perms  Size      Offset    File")
        print("-" * 80)
        for page in pages:
            perms = ""
            perms += "r" if page.read else "-"
            perms += "w" if page.write else "-"
            perms += "x" if page.execute else "-"
            size = page.end - page.start
            offset = f"{page.offset:#x}" if page.offset is not None else "-"
            objfile = page.objfile if page.objfile else ""
            print(
                f"{page.start:#018x}-{page.end:#018x} {perms}  {size:#010x}  {offset:8}  {objfile}"
            )

    return pages


def aliases() -> None:
    """Print help for all available shortcuts."""
    help_text = """
╔═══════════════════════════════════════════════════════════════════════════════╗
║                     Pwndbg IPython Shortcuts Reference                       ║
╚═══════════════════════════════════════════════════════════════════════════════╝

MEMORY OPERATIONS:
  mr(addr, count=0x40)                 Read memory bytes
  mw(addr, data)                       Write memory
  hd(addr, count=0x40)                 Hexdump memory
  ms(pattern, start, end, limit, show) Search memory for bytes

REGISTER OPERATIONS:
  regs.<name>                          Read register value
  rr(name)                             Read register value
  rw(name, value)                      Write register value

VIRTUAL MEMORY:
  vm(show=True)                        Show virtual memory mappings

For more info: type help(function_name), e.g., help(mr)
"""
    print(help_text)


def get_ipi_namespace() -> dict[str, Any]:
    """Get dictionary of all helpers to inject into IPython namespace.

    Returns:
        Dictionary mapping names to helper functions
    """
    return {
        "mr": mr,
        "mw": mw,
        "hd": hd,
        "ms": ms,
        "rr": rr,
        "rw": rw,
        "vm": vm,
        "aliases": aliases,
    }


def get_banner() -> str:
    """Get banner text to display when entering ipi.

    Returns:
        Banner string
    """
    is_dead_warn = "Note: `process is not alive now`\n" if not pwndbg.aglib.proc.alive() else ""
    return color_tip(
        "Shortcuts: read/write regs: `rr(name)`, `rw(name, val)` | memory: `mr(addr, count)`, `wr(addr, count)`\n"
        "           hexdump: `hd(addr, count)`                 | vmmap: `vm()`\n"
        + is_dead_warn
        + "Use `aliases()` for help"
    )



from __future__ import annotations

import pwndbg.aglib.memory
import pwndbg.aglib.symbol
import pwndbg.aglib.typeinfo

# adapted from jemalloc source 5.3.0
LG_VADDR = 48
LG_PAGE = 12
# https://github.com/jemalloc/jemalloc/blob/a25b9b8ba91881964be3083db349991bbbbf1661/include/jemalloc/internal/jemalloc_internal_types.h#L42
MALLOCX_ARENA_BITS = 12
# https://github.com/jemalloc/jemalloc/blob/a25b9b8ba91881964be3083db349991bbbbf1661/include/jemalloc/jemalloc_defs.h.in#L51
LG_SIZEOF_PTR = 3

RTREE_NHIB = (1 << (LG_SIZEOF_PTR + 3)) - LG_VADDR  # Number of high insignificant bits
RTREE_NLIB = LG_PAGE  # Number of low insigificant bits
RTREE_NSB = LG_VADDR - RTREE_NLIB  # Number of significant bits

# Number of levels in radix tree
if RTREE_NSB <= 10:
    RTREE_HEIGHT = 1
elif RTREE_NSB <= 36:
    RTREE_HEIGHT = 2
elif RTREE_NSB <= 52:
    RTREE_HEIGHT = 3
else:
    raise ValueError("Unsupported number of significant virtual address bits")


# TODO: RTREE_LEAF_COMPACT should be enabled otherwise rtree_leaf_elm_s would change

# TODO: Move to relevant place
# https://github.com/jemalloc/jemalloc/blob/a25b9b8ba91881964be3083db349991bbbbf1661/include/jemalloc/internal/edata.h#L145


def mask(current_field_width, current_field_shift):
    return ((1 << current_field_width) - 1) << current_field_shift


# For size class related explanation and calculations, refer to https://github.com/jemalloc/jemalloc/blob/a25b9b8ba91881964be3083db349991bbbbf1661/include/jemalloc/internal/sc.h#L8

LG_QUANTUM = 4  # LG_QUANTUM ensures correct platform alignment and necessary to ensure we never return improperly aligned memory

SC_LG_TINY_MIN = 3
SC_NTINY = (
    LG_QUANTUM - SC_LG_TINY_MIN
)  # Number of tiny size classes for alloations smaller than (1 << LG_QUANTUM)

# Size classes
SC_LG_NGROUP = 2  # Number of size classes group
SC_NGROUP = (
    1 << SC_LG_NGROUP
)  # Number of size classes in each group, equally spaced in the range, so that * each one covers allocations for base / SC_NGROUP possible allocation sizes
SC_NPSEUDO = SC_NGROUP
SC_PTR_BITS = (1 << LG_SIZEOF_PTR) * 8
SC_LG_BASE_MAX = SC_PTR_BITS - 2
SC_LG_FIRST_REGULAR_BASE = LG_QUANTUM + SC_LG_NGROUP
SC_NREGULAR = SC_NGROUP * (SC_LG_BASE_MAX - SC_LG_FIRST_REGULAR_BASE + 1) - 1

SC_NSIZES = SC_NTINY + SC_NPSEUDO + SC_NREGULAR

SC_LG_SLAB_MAXREGS = LG_PAGE - SC_LG_TINY_MIN


# Source: https://github.com/jemalloc/jemalloc/blob/dev/include/jemalloc/internal/bit_util.h#L400-L419
def lg_floor_1(x):
    return 0


def lg_floor_2(x):
    return lg_floor_1(x) if x < (1 << 1) else 1 + lg_floor_1(x >> 1)


def lg_floor_4(x):
    return lg_floor_2(x) if x < (1 << 2) else 2 + lg_floor_2(x >> 2)


def lg_floor_8(x):
    return lg_floor_4(x) if x < (1 << 4) else 4 + lg_floor_4(x >> 4)


def lg_floor_16(x):
    return lg_floor_8(x) if x < (1 << 8) else 8 + lg_floor_8(x >> 8)


def lg_floor_32(x):
    return lg_floor_16(x) if x < (1 << 16) else 16 + lg_floor_16(x >> 16)


def lg_floor_64(x):
    return lg_floor_32(x) if x < (1 << 32) else 32 + lg_floor_32(x >> 32)


def lg_floor(x):
    return lg_floor_32(x) if LG_SIZEOF_PTR == 2 else lg_floor_64(x)


def lg_ceil(x):
    return lg_floor(x) + (0 if (x & (x - 1)) == 0 else 1)


# Arena width and mask definitions
EDATA_BITS_ARENA_WIDTH = MALLOCX_ARENA_BITS
EDATA_BITS_ARENA_SHIFT = 0
EDATA_BITS_ARENA_MASK = mask(EDATA_BITS_ARENA_WIDTH, EDATA_BITS_ARENA_SHIFT)

# Slab width and mask definitions
EDATA_BITS_SLAB_WIDTH = 1
EDATA_BITS_SLAB_SHIFT = EDATA_BITS_ARENA_WIDTH + EDATA_BITS_ARENA_SHIFT
EDATA_BITS_SLAB_MASK = mask(EDATA_BITS_SLAB_WIDTH, EDATA_BITS_SLAB_SHIFT)

# Committed width and mask definitions
EDATA_BITS_COMMITTED_WIDTH = 1
EDATA_BITS_COMMITTED_SHIFT = EDATA_BITS_SLAB_WIDTH + EDATA_BITS_SLAB_SHIFT
EDATA_BITS_COMMITTED_MASK = mask(EDATA_BITS_COMMITTED_WIDTH, EDATA_BITS_COMMITTED_SHIFT)

# PAI width and mask definitions
EDATA_BITS_PAI_WIDTH = 1
EDATA_BITS_PAI_SHIFT = EDATA_BITS_COMMITTED_WIDTH + EDATA_BITS_COMMITTED_SHIFT
EDATA_BITS_PAI_MASK = mask(EDATA_BITS_PAI_WIDTH, EDATA_BITS_PAI_SHIFT)

# Zeroed width and mask definitions
EDATA_BITS_ZEROED_WIDTH = 1
EDATA_BITS_ZEROED_SHIFT = EDATA_BITS_PAI_WIDTH + EDATA_BITS_PAI_SHIFT
EDATA_BITS_ZEROED_MASK = mask(EDATA_BITS_ZEROED_WIDTH, EDATA_BITS_ZEROED_SHIFT)

# Guarded width and mask definitions
EDATA_BITS_GUARDED_WIDTH = 1
EDATA_BITS_GUARDED_SHIFT = EDATA_BITS_ZEROED_WIDTH + EDATA_BITS_ZEROED_SHIFT
EDATA_BITS_GUARDED_MASK = mask(EDATA_BITS_GUARDED_WIDTH, EDATA_BITS_GUARDED_SHIFT)

# State width and mask definitions
EDATA_BITS_STATE_WIDTH = 3
EDATA_BITS_STATE_SHIFT = EDATA_BITS_GUARDED_WIDTH + EDATA_BITS_GUARDED_SHIFT
EDATA_BITS_STATE_MASK = mask(EDATA_BITS_STATE_WIDTH, EDATA_BITS_STATE_SHIFT)

EDATA_BITS_SZIND_WIDTH = lg_ceil(SC_NSIZES)
EDATA_BITS_SZIND_SHIFT = EDATA_BITS_STATE_WIDTH + EDATA_BITS_STATE_SHIFT
EDATA_BITS_SZIND_MASK = mask(EDATA_BITS_SZIND_WIDTH, EDATA_BITS_SZIND_SHIFT)

# Nfree width and mask definitions
EDATA_BITS_NFREE_WIDTH = SC_LG_SLAB_MAXREGS + 1
EDATA_BITS_NFREE_SHIFT = EDATA_BITS_SZIND_WIDTH + EDATA_BITS_SZIND_SHIFT
EDATA_BITS_NFREE_MASK = mask(EDATA_BITS_NFREE_WIDTH, EDATA_BITS_NFREE_SHIFT)

# Binshard width and mask definitions
EDATA_BITS_BINSHARD_WIDTH = 6
EDATA_BITS_BINSHARD_SHIFT = EDATA_BITS_NFREE_WIDTH + EDATA_BITS_NFREE_SHIFT
EDATA_BITS_BINSHARD_MASK = mask(EDATA_BITS_BINSHARD_WIDTH, EDATA_BITS_BINSHARD_SHIFT)

# Is head width and mask definitions
EDATA_BITS_IS_HEAD_WIDTH = 1
EDATA_BITS_IS_HEAD_SHIFT = EDATA_BITS_BINSHARD_WIDTH + EDATA_BITS_BINSHARD_SHIFT
EDATA_BITS_IS_HEAD_MASK = mask(EDATA_BITS_IS_HEAD_WIDTH, EDATA_BITS_IS_HEAD_SHIFT)

# In RTree, Each level distinguishes a certain number of bits from the key, which helps in narrowing down the search space
# bits: how many bits have been used at that particular level (Number of key bits distinguished by this level)
# cumbits: how many bits in total have been used up to that level (Cumulative number of key bits distinguished by traversing to corresponding tree level)
rtree_levels = [
    # for height == 1
    [{"bits": RTREE_NSB, "cumbits": RTREE_NHIB + RTREE_NSB}],
    # for height == 2
    [
        {"bits": RTREE_NSB // 2, "cumbits": RTREE_NHIB + RTREE_NSB // 2},
        {"bits": RTREE_NSB // 2 + RTREE_NSB % 2, "cumbits": RTREE_NHIB + RTREE_NSB},
    ],
    # for height == 3
    [
        {"bits": RTREE_NSB // 3, "cumbits": RTREE_NHIB + RTREE_NSB // 3},
        {
            "bits": RTREE_NSB // 3 + RTREE_NSB % 3 // 2,
            "cumbits": RTREE_NHIB + RTREE_NSB // 3 * 2 + RTREE_NSB % 3 // 2,
        },
        {
            "bits": RTREE_NSB // 3 + RTREE_NSB % 3 - RTREE_NSB % 3 // 2,
            "cumbits": RTREE_NHIB + RTREE_NSB,
        },
    ],
]


class RTree:
    """
    RTree is used by jemalloc to keep track of extents that are allocated by jemalloc.
    Since extent data is not stored in a doubly linked list, rtree is used to find the extent belonging to a pointer that is being freed.
    Implementation of rtree is similar to Linux Radix tree: https://lwn.net/Articles/175432/
    """

    # TODO: Check rtee_ctx cache in
    # tsd_nominal_tsds.qlh_first.cant_access_tsd_items_directly_use_a_getter_or_setter_rtree_ctx.cache
    def __init__(self, addr: int) -> None:
        self._addr = addr

        rtree_s = pwndbg.aglib.typeinfo.load("struct rtree_s")
        if rtree_s is None:
            raise pwndbg.dbg_mod.Error("rtree_s type not found")

        # self._Value = pwndbg.aglib.memory.poi(emap_s, self._addr)

        # self._Value = pwndbg.aglib.memory.fetch_struct_as_dictionary(
        #     "rtree_s", self._addr, include_only_fields={"root"}
        # )
        # pwndbg.aglib.memory
        self._Value = pwndbg.aglib.memory.get_typed_pointer_value("struct rtree_s", self._addr)

        self._extents = None

    @staticmethod
    def get_rtree() -> RTree:
        addr = pwndbg.aglib.symbol.lookup_symbol_addr("je_arena_emap_global")
        if addr is None:
            raise pwndbg.dbg_mod.Error("Required je_arena_emap_global symbol not found")
        return RTree(addr)

    @property
    def root(self):
        return self._Value["root"]

    # from include/jemalloc/internal/rtree.h
    # converted implementation of rtree_leafkey
    def __rtree_leaf_maskbits(self, level):
        ptrbits = 1 << (LG_SIZEOF_PTR + 3)
        # print("ptrbits: ", ptrbits, bin(ptrbits))
        cumbits = (
            rtree_levels[RTREE_HEIGHT - 1][level - 1]["cumbits"]
            - rtree_levels[RTREE_HEIGHT - 1][level - 1]["bits"]
        )
        # print("cumbits: ", cumbits, bin(cumbits))
        return ptrbits - cumbits

    # Can be used to lookup key quickly in cache
    def __rtree_leafkey(self, key: int, level: int) -> int:
        mask = ~((1 << self.__rtree_leaf_maskbits(level)) - 1)
        # print("mask: ", mask, bin(mask))
        return key & mask

    def __subkey(self, key: int, level: int) -> int:
        """
        Return a portion of the key that is used to find the node/leaf in the rtree at a specific level.
        Source: https://github.com/jemalloc/jemalloc/blob/5b72ac098abce464add567869d082f2097bd59a2/include/jemalloc/internal/rtree.h#L161
        """

        ptrbits = 1 << (LG_SIZEOF_PTR + 3)
        cumbits = rtree_levels[RTREE_HEIGHT - 1][level - 1]["cumbits"]
        shiftbits = ptrbits - cumbits
        maskbits = rtree_levels[RTREE_HEIGHT - 1][level - 1]["bits"]
        mask = (1 << maskbits) - 1

        return (key >> shiftbits) & mask

    @staticmethod
    def __alignment_addr2base(addr, alignment=64):
        return addr - (addr - (addr & (~(alignment - 1))))

    def lookup_hard(self, key: int):
        """
        Lookup the key in the rtree and return the value.

        How it works:
        - Jemalloc stores the extent address in the rtree as a node and to find a specific node we need a address key.
        """
        rtree_node_elm_s = pwndbg.aglib.typeinfo.load("struct rtree_node_elm_s")
        if rtree_node_elm_s is None:
            raise pwndbg.dbg_mod.Error("rtree_node_elm_s type not found")
        rtree_leaf_elm_s = pwndbg.aglib.typeinfo.load("struct rtree_leaf_elm_s")
        if rtree_leaf_elm_s is None:
            raise pwndbg.dbg_mod.Error("rtree_leaf_elm_s type not found")

        # Credits: 盏一's jegdb
        # https://web.archive.org/web/20221114090949/https://github.com/hidva/hidva.github.io/blob/dev/_drafts/jegdb.py

        # For subkey 0
        subkey = self.__subkey(key, 1)

        addr = int(self.root.address) + subkey * rtree_node_elm_s.sizeof
        fetched_struct = pwndbg.aglib.memory.get_typed_pointer_value(
            "struct rtree_node_elm_s", addr
        )
        child_repr = int(fetched_struct["child"]["repr"])

        # on node element, child contains the bits with which we can find another node or leaf element
        if child_repr == 0:
            return None

        # For subkey 1
        subkey = self.__subkey(key, 2)
        addr = child_repr + subkey * rtree_leaf_elm_s.sizeof
        fetched_struct = pwndbg.aglib.memory.get_typed_pointer_value(
            "struct rtree_leaf_elm_s", addr
        )

        # On leaf element, le_bits contains the virtual memory address bits so we can use it to find the extent address
        val = int(fetched_struct["le_bits"]["repr"])
        if val == 0:
            return None

        # In this function, we are trying to find the extent address given the address of memory block
        # that this extent is managing (which is represented by edata->e_addr in the extent structure)

        # e_addr is 64 bits but
        # e_addr is also page (4096) aligned which means last 12 bits are zero and therefore unused
        # In rtree, each layer can be accessed using bits 0-16, 17-33 and 34-51
        # When height of rtree is 3, level 1 can be accessed using bits 0-16, and so on for level 2 and 3
        # When the height is 2, 0-15 bits are unused and level 1 can be accessed using bits 16-33 and level 2 using 34-51

        ls = (val << RTREE_NHIB) & ((2**64) - 1)
        ptr = ((ls >> RTREE_NHIB) >> 1) << 1

        if ptr == 0:
            return None

        # return Extent(ptr)
        extent = Extent(ptr)
        if extent.size == 0:
            ptr = RTree.__alignment_addr2base(ptr)
            extent_tmp = Extent(ptr)
            if extent_tmp.size != 0:
                return extent_tmp

        return extent

    @property
    def extents(self):
        # NOTE: Generating whole extents list is slow as it requires parsing whole rtree

        if self._extents is None:  # TODO: handling cache on extents changes
            self._extents = []
            try:
                root = self.root
                last_addr = None
                extent_addresses = []

                rtree_node_elm_s = pwndbg.aglib.typeinfo.load("struct rtree_node_elm_s")
                if rtree_node_elm_s is None:
                    raise pwndbg.dbg_mod.Error("rtree_node_elm_s type not found")
                rtree_leaf_elm_s = pwndbg.aglib.typeinfo.load("struct rtree_leaf_elm_s")
                if rtree_leaf_elm_s is None:
                    raise pwndbg.dbg_mod.Error("rtree_leaf_elm_s type not found")

                max_subkeys = 1 << rtree_levels[RTREE_HEIGHT - 1][0]["bits"]
                # print("max_subkeys: ", max_subkeys)

                for i in range(max_subkeys):
                    node_address = int(root.address) + i * rtree_node_elm_s.sizeof
                    # node = pwndbg.aglib.memory.poi(rtree_node_elm_s, node)
                    fetched_struct = pwndbg.aglib.memory.get_typed_pointer_value(
                        rtree_node_elm_s, node_address
                    )
                    leaf0 = int(fetched_struct["child"]["repr"])
                    if leaf0 == 0:
                        continue

                    # print(hex(leaf0))
                    # print("leaf0: ", leaf0)

                    # level 1
                    for j in range(max_subkeys):
                        leaf_address = leaf0 + j * rtree_leaf_elm_s.sizeof
                        # leaf = pwndbg.aglib.memory.poi(rtree_leaf_elm_s, leaf)
                        fetched_struct = pwndbg.aglib.memory.get_typed_pointer_value(
                            rtree_leaf_elm_s, leaf_address
                        )
                        val = int(fetched_struct["le_bits"]["repr"])
                        if val == 0:
                            continue

                        # print("leaf: ", hex(leaf_address))
                        # print(j, leaf)

                        ls = (val << RTREE_NHIB) & ((2**64) - 1)
                        ptr = ((ls >> RTREE_NHIB) >> 1) << 1

                        if ptr == 0 or ptr == last_addr:
                            continue

                        last_addr = ptr

                        extent = Extent(ptr)

                        if extent.extent_address in extent_addresses:
                            continue

                        extent_addresses.append(extent.extent_address)

                        # during initializations, addresses may get some alignment
                        # lets check if size makes sense, otherwise do page alignment and check if again
                        # TODO: better way to do this
                        extent_tmp = extent
                        if extent.size == 0:
                            ptr = RTree.__alignment_addr2base(int(ptr))
                            extent_tmp = Extent(ptr)
                            if extent_tmp.size != 0:
                                self._extents.append(extent_tmp)
                                continue

                        self._extents.append(extent_tmp)

            except pwndbg.dbg_mod.Error:
                pass

        return self._extents


class Extent:
    """
    Concept of extent (edata) is similar to chunk in glibc malloc but allocation algorithm differs a lot.
    - Extents are used to manage memory blocks (including jemalloc metadata) where extents sizes can vary but each block is always a multiple of the page size.
    - jemalloc will either allocate one large class request or multiple small class request (called slab) depending on request size.
    - Unlike chunks in glibc malloc, extents are not doubly linked list but are managed using rtree.
    - This tree is mostly used during deallocation to find the extent belonging to a pointer that is being freed.
    - Extents are also not stored as a header structure but externally (therefore extent metadata and actually mapped data may be very far apart).
    """

    def __init__(self, addr: int) -> None:
        self._addr = addr

        # fetch_struct_as_dictionary does not support union currently
        self._Value = pwndbg.aglib.memory.get_typed_pointer_value("struct edata_s", self._addr)

        self._bitfields = None

    @property
    def size(self):
        """
        May be larger in case of large size class allocation when cache_oblivious is enabled.
        """
        # return self._Value["e_size_esn"]
        return (int(self._Value["e_size_esn"]) >> LG_PAGE) << LG_PAGE

    @property
    def extent_address(self) -> int:
        """
        Address of the extent data structure (not the actual memory).
        """
        return self._addr

    @property
    def allocated_address(self) -> int:
        """
        Starting address of allocated memory
        cache-oblivious large allocation alignment:
            When a large class allocation is made, jemalloc selects the closest size class that can fit the request and allocates that size + 4 KiB (0x1000).
            However, the pointer returned to user is randomized between the 'base' and 'base + 4 KiB' (0x1000) range.
            Source code: https://github.com/jemalloc/jemalloc/blob/a25b9b8ba91881964be3083db349991bbbbf1661/include/jemalloc/internal/arena_inlines_b.h#L505
        """
        return int(self._Value["e_addr"])

    @property
    def bsize(self) -> int:
        return int(self._Value["e_bsize"])

    @property
    def bits(self) -> int:
        return int(self._Value["e_bits"])

    @property
    def bitfields(self) -> dict[str, int]:
        """
        Extract bitfields

        arena_ind: Arena from which this extent came, or all 1 bits if unassociated.
        slab: The slab flag indicates whether the extent is used for a slab of small regions. This helps differentiate small size classes, and it indicates whether interior pointers can be looked up via iealloc().
        committed: The committed flag indicates whether physical memory is committed to the extent, whether explicitly or implicitly as on a system that overcommits and satisfies physical memory needs on demand via soft page faults.
        pai: The pai flag is an extent_pai_t.
        zeroed: The zeroed flag is used by extent recycling code to track whether memory is zero-filled.
        guarded: The guarded flag is used by the sanitizer to track whether the extent has page guards around it.
        state: The state flag is an extent_state_t.
        szind: The szind flag indicates usable size class index for allocations residing in this extent, regardless of whether the extent is a slab. Extent size and usable size often differ even for non-slabs, either due to sz_large_pad or promotion of sampled small regions.
        nfree: Number of free regions in slab.
        bin_shard: The shard of the bin from which this extent came.
        """

        if self._bitfields is None:
            self._bitfields = {
                "arena_ind": (self.bits & EDATA_BITS_ARENA_MASK) >> EDATA_BITS_ARENA_SHIFT,
                "slab": (self.bits & EDATA_BITS_SLAB_MASK) >> EDATA_BITS_SLAB_SHIFT,
                "committed": (self.bits & EDATA_BITS_COMMITTED_MASK) >> EDATA_BITS_COMMITTED_SHIFT,
                "pai": (self.bits & EDATA_BITS_PAI_MASK) >> EDATA_BITS_PAI_SHIFT,
                "zeroed": (self.bits & EDATA_BITS_ZEROED_MASK) >> EDATA_BITS_ZEROED_SHIFT,
                "guarded": (self.bits & EDATA_BITS_GUARDED_MASK) >> EDATA_BITS_GUARDED_SHIFT,
                "state": (self.bits & EDATA_BITS_STATE_MASK) >> EDATA_BITS_STATE_SHIFT,
                "szind": (self.bits & EDATA_BITS_SZIND_MASK) >> EDATA_BITS_SZIND_SHIFT,
                "nfree": (self.bits & EDATA_BITS_NFREE_MASK) >> EDATA_BITS_NFREE_SHIFT,
                "bin_shard": (self.bits & EDATA_BITS_BINSHARD_MASK) >> EDATA_BITS_BINSHARD_SHIFT,
            }

        return self._bitfields

    @property
    def state_name(self) -> str:
        state_mapping = ["Active", "Dirty", "Muzzy", "Retained"]

        return state_mapping[self.bitfields["state"]]

    @property
    def has_slab(self) -> bool:
        """
        Returns True if the extent is used for small size classes.
        Reference for size in Table 1 at https://jemalloc.net/jemalloc.3.html
        At time of writing, allocations <= 0x3800 are considered as small allocations and has slabs.
        """
        return self.bitfields["slab"] != 0

    @property
    def is_free(self) -> bool:
        """
        Returns True if the extent is free.
        """
        pass

    @property
    def pai(self) -> str:
        """
        Page Allocator Interface
        """
        if self.bitfields["pai"] == 0:
            return "PAC"  # Page for extent
        return "HPA"  # Huge Page



"""
Ignoring a breakpoint
"""

from __future__ import annotations

import argparse

import gdb

import pwndbg.commands
from pwndbg.color import message
from pwndbg.commands import CommandCategory

parser = argparse.ArgumentParser(
    description="""Set ignore-count of breakpoint number N to COUNT.

While the ignore count is positive, execution will not stop on the breakpoint.

By default, if `N' is ommitted, the last breakpoint (i.e. greatest breakpoint number) will be used.""",
)
parser.add_argument(
    "bpnum", metavar="N", type=int, default=None, nargs="?", help="The breakpoint number N."
)
parser.add_argument("count", metavar="COUNT", type=int, help="The number to set COUNT.")


@pwndbg.commands.Command(parser, category=CommandCategory.BREAKPOINT)
def ignore(bpnum, count) -> None:
    bps = gdb.breakpoints()

    if not bps:
        print(message.error("No breakpoints set."))
        return

    if bpnum is None:
        bp = max(bps, key=lambda bp: bp.number)
    else:
        bp = next((bp for bp in bps if bp.number == bpnum), None)

        if bp is None:
            print(message.error(f"No breakpoint number {bpnum}."))
            return

    count = max(0, int(count))
    bp.ignore_count = count
    print(f"Will ignore next {count} crossings of breakpoint {bp.number}.")



from __future__ import annotations

import pwndbg.aglib.kernel


def find_dmabuf_offsets(dmabuf) -> tuple[int, int, int]:
    MAX = 0x30
    sg_table_off, exp_name_off, list_node_off = None, None, None
    ptrsize = pwndbg.aglib.arch.ptrsize
    heap_buffer = pwndbg.aglib.memory.read_pointer_width(dmabuf + 2 * ptrsize)
    for i in range(1, MAX):
        # see recover_dmabuf_typeinfo (struct dma_buf) for an explanation
        # this loop is searching the `size` field from `list_node`
        size = pwndbg.aglib.memory.read_pointer_width(dmabuf - (i + 5) * ptrsize)
        file = pwndbg.aglib.memory.read_pointer_width(dmabuf - (i + 4) * ptrsize)
        attachments_prev = pwndbg.aglib.memory.read_pointer_width(dmabuf - (i + 3) * ptrsize)
        attachments_next = pwndbg.aglib.memory.read_pointer_width(dmabuf - (i + 2) * ptrsize)
        ops = pwndbg.aglib.memory.read_pointer_width(dmabuf - (i + 1) * ptrsize)
        vmapping_counter = pwndbg.aglib.memory.read_pointer_width(dmabuf - i * ptrsize)
        if pwndbg.aglib.memory.is_kernel(size):
            continue
        if not pwndbg.aglib.memory.is_kernel(file):
            continue
        if not (
            pwndbg.aglib.memory.is_kernel(attachments_next)
            and pwndbg.aglib.memory.is_kernel(attachments_prev)
        ):
            continue
        if not pwndbg.aglib.memory.is_kernel(ops):
            continue
        if pwndbg.aglib.memory.is_kernel(vmapping_counter):
            continue
        # (i + 5) * ptrsize is the distance from the `size` to `list_node`
        list_node_off = (i + 5) * ptrsize
        break
    assert list_node_off is not None, "cannot determine the offset of list_node"
    dmabuf -= list_node_off
    for i in range(5, MAX):
        ptr = pwndbg.aglib.memory.read_pointer_width(dmabuf + i * ptrsize)
        try:
            if len(pwndbg.aglib.memory.string(ptr).decode()) == 0:
                continue
        except Exception:
            continue
        exp_name_off = i * ptrsize
        break
    assert exp_name_off is not None, "cannot determine the offset of exp_name"
    sz = pwndbg.aglib.memory.read_pointer_width(dmabuf)
    for i in range(MAX):
        if pwndbg.aglib.memory.read_pointer_width(heap_buffer + i * ptrsize) == sz:
            sg_table_off = (i + 1) * ptrsize
            break
    assert sg_table_off is not None, "cannot determine the offset of sg_table"
    return sg_table_off, exp_name_off, list_node_off


@pwndbg.aglib.kernel.typeinfo_recovery("struct dma_buf")
def recover_dmabuf_typeinfo(first_dmabuf: int) -> str:
    # reaching here means priv exists
    sg_table_off, exp_name_off, list_node_off = find_dmabuf_offsets(first_dmabuf)
    result = pwndbg.aglib.kernel.symbol.COMMON_TYPES
    result += f"""
    typedef unsigned long dma_addr_t;
    struct scatterlist {{
        unsigned long	page_link; // either to a page or scatterlist
        unsigned int	offset;
        unsigned int	length;
        dma_addr_t	dma_address;
    /*** potentially has 8 more bytes
#ifdef CONFIG_NEED_SG_DMA_LENGTH
        unsigned int	dma_length;
#endif
#ifdef CONFIG_NEED_SG_DMA_FLAGS
        unsigned int    dma_flags;
#endif
    ***/
    }};
    struct sg_table {{
        struct scatterlist *sgl;	/* the list */
        unsigned int nents;		/* number of mapped entries */
        unsigned int orig_nents;	/* original size of list */
    }};
    struct system_heap_buffer {{
        char _pad[{sg_table_off}];
        struct sg_table sg_table;
        /* rest of the fields are irrelevant */
    }};
    struct dma_buf {{
        size_t size; // (i + 5) is here
        void *file;
        struct list_head attachments;
        void *ops; // const struct dma_buf_ops *
        unsigned vmapping_counter;
        char _pad1[{exp_name_off - pwndbg.aglib.arch.ptrsize * 6}];
        const char *exp_name;
        const char *name;
        char _pad2[{list_node_off - exp_name_off - pwndbg.aglib.arch.ptrsize * 2}];
        struct list_head list_node;
        struct system_heap_buffer *priv; // treating the voidptr as system_heap_buffer
        /* rest of the fields are irrelevant */
    }};
    """
    return result



from __future__ import annotations

import argparse
import re
import tempfile
from collections.abc import Iterator

import pwndbg.aglib
import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.proc
import pwndbg.aglib.vmmap
import pwndbg.color.disasm
import pwndbg.color.memory
import pwndbg.color.message as message
import pwndbg.commands
import pwndbg.dbg_mod
import pwndbg.dintegration
import pwndbg.lib.memory
from pwndbg.aglib.disasm.disassembly import get_disassembler
from pwndbg.commands import CommandCategory


class RawMemoryBinary:
    def __init__(self, options, start_addr: int):
        self.start_addr = start_addr
        self.__fileName = options.binary
        self.__rawBinary = None
        self.cs = get_disassembler(pwndbg.aglib.arch.get_capstone_constants(pwndbg.aglib.regs.pc))

        with open(self.__fileName, "rb") as fp:
            self.__rawBinary = fp.read()

    def getBinary(self):
        return self

    def getFileName(self):
        return self.__fileName

    def getRawBinary(self):
        return self.__rawBinary

    def getEntryPoint(self):
        raise NotImplementedError()

    def getExecSections(self):
        return [
            {
                "name": "raw",
                "offset": 0,
                "size": len(self.__rawBinary),
                "vaddr": self.start_addr,
                "opcodes": bytes(self.__rawBinary),
            }
        ]

    def getDataSections(self):
        raise NotImplementedError()

    def getArch(self):
        return self.cs.arch

    def getArchMode(self):
        return self.cs.mode

    def getEndian(self):
        # Already returned in `getArchMode` func
        return 0

    def getFormat(self):
        return "Raw"


def _rop(
    file_path: str,
    grep: str | None,
    argument: list[str],
    start_addr: int | None = None,
    symbols: bool = False,
    plain: bool = False,
) -> bool:
    import contextlib
    from io import StringIO

    from ropgadget.args import Args
    from ropgadget.core import Core

    stderr = StringIO()

    with contextlib.redirect_stderr(stderr):
        try:
            args = Args(
                arguments=[
                    "--binary",
                    file_path,
                    *argument,
                ]
            )
        except SystemExit as e:  # ropgadget runs argparse which calls sys.exit
            if e.code == 2:  # invalid args
                full = stderr.getvalue()
                print(
                    message.error(full.splitlines()[-1].removeprefix(": error: "))
                )  # we skip the usage block, and only print the error
            return False

    options = args.getArgs()
    c = Core(options)

    if start_addr is not None:
        # HACK: to load from our class
        c._Core__binary = RawMemoryBinary(options, start_addr=start_addr)
    else:
        c.do_binary(file_path, silent=True)

    # Find gadgets
    c.do_load(0, silent=True)

    if symbols:
        decomp_stack_vars: dict[int, str] = pwndbg.dintegration.manager.get_stack_var_dict_all()
    else:
        decomp_stack_vars = {}

    print("Gadgets information\n============================================================")
    for gadget in c.gadgets():
        insts = gadget.get("gadget", "")
        if not insts:
            continue

        vaddr = gadget["vaddr"]

        n_insts = insts.count(";") + 1
        enhanced_insts = pwndbg.aglib.disasm.disassembly.get(
            vaddr, n_insts, enhance=not plain, padding=0
        )
        insts_str = " ; ".join(ins.asm_string for ins in enhanced_insts)

        if symbols:
            out = f"{pwndbg.color.memory.get_address_and_symbol(vaddr, decomp_stack_vars)}: {insts_str}"
        else:
            out = f"{pwndbg.color.memory.get(vaddr)}: {insts_str}"

        plain_out = pwndbg.color.strip(out)

        if grep:
            # grep search
            if not re.search(grep, insts) and not re.search(grep, plain_out):
                continue

        print(plain_out if plain else out)

    print(f"\nUnique gadgets found: {len(c.gadgets())}")
    return True


def split_range_to_chunks(
    range_start: int, range_end: int, chunk_size: int = 10 * 1024 * 1024
) -> Iterator[tuple[int, int, int, int]]:
    total_parts = ((range_end - range_start) + chunk_size - 1) // chunk_size

    for current_part, range_start_chunk in enumerate(range(range_start, range_end, chunk_size), 1):
        range_end_chunk = min(range_start_chunk + chunk_size, range_end)
        range_size = range_end_chunk - range_start_chunk

        yield (
            range_start_chunk,
            range_size,
            current_part,
            total_parts,
        )


def parse_size(size_str: str) -> int:
    unit_multipliers = {
        "b": 1,
        "kb": 1024,
        "mb": 1024**2,
        "gb": 1024**3,
        "tb": 1024**4,
        "kib": 1024,
        "mib": 1024**2,
        "gib": 1024**3,
        "tib": 1024**4,
    }
    size_str = size_str.strip().lower()

    match = re.match(r"(\d+)\s*(b|kb|mb|gb|tb|kib|mib|gib|tib)", size_str)
    if not match:
        raise ValueError(f"Invalid size string: {size_str}")

    value = int(match.group(1))
    unit = match.group(2)
    return value * unit_multipliers[unit]


def iterate_over_pages(mem_limit: int) -> Iterator[tuple[str, pwndbg.lib.memory.Page | None]]:
    if not pwndbg.aglib.proc.alive():
        yield pwndbg.aglib.proc.exe(), None
        return

    proc = pwndbg.dbg.selected_inferior()
    for page in pwndbg.aglib.vmmap.get_memory_map().ranges():
        if not page.execute:
            continue

        print(message.info(f"Searching in {hex(page.start)} {hex(page.end)} {page.objfile}"))
        if page.memsz > mem_limit:
            print(
                message.hint(
                    "WARNING: The memory page size is too large to dump.\n"
                    "WARNING: Parsing this large memory page might take an excessive amount of time...\n"
                    "WARNING: To process larger pages, increase the `--memlimit` parameter (e.g., `--memlimit 100MB`)."
                )
            )
            continue

        with tempfile.NamedTemporaryFile(mode="a+b") as fmem:
            try:
                for start, size, progress_cur, progress_max in split_range_to_chunks(
                    page.start, page.end
                ):
                    if progress_max > 1:
                        print(message.hint(f"Dumping memory... {progress_cur} / {progress_max}"))

                    mem_data = proc.read_memory(address=start, size=size)
                    fmem.write(mem_data)
            except pwndbg.dbg_mod.Error as e:
                print(message.error(f"WARNING: failed to read page: {e}"))
                continue

            fmem.flush()
            yield fmem.name, page


parser = argparse.ArgumentParser(
    description="Dump ROP gadgets with Jon Salwan's ROPgadget tool.",
)
parser.add_argument("--grep", type=str, help="String to grep the output for")
parser.add_argument(
    "--memlimit", type=str, default="50MB", help="Maximum size of memory pages to scan"
)
parser.add_argument(
    "--symbols", action="store_true", help="Show symbols for/of gadgets (if there are any)"
)
parser.add_argument("--plain", action="store_true", help="Plain output (no highlighting)")
parser.add_argument("arguments", nargs="*", type=str, help="Arguments to pass to ROPgadget")


@pwndbg.commands.Command(
    parser,
    aliases=["ropgadget"],
    category=CommandCategory.INTEGRATIONS,
    examples="""
pwndbg> rop --grep 'pop rdi' -- --nojop
Gadgets information
============================================================
[...]
0x0007dce8 : pop rdi ; or dword ptr [rax], eax ; add rsp, 0x28 ; ret
0x0007d838 : pop rdi ; or dword ptr [rax], eax ; je 0x7d840 ; ret
0x0005b13d : pop rdi ; pop rbp ; ret
0x0001ee23 : pop rdi ; ret

Unique gadgets found: 8514
    """,
)
@pwndbg.commands.OnlyWithFile
def rop(grep: str | None, memlimit: str, symbols: bool, plain: bool, arguments: list[str]) -> None:
    memlimit = parse_size(memlimit)

    for file_path, page in iterate_over_pages(memlimit):
        should_continue = _rop(
            file_path,
            grep,
            arguments,
            start_addr=page.start if page else None,
            symbols=symbols,
            plain=plain,
        )
        if not should_continue:
            break



from __future__ import annotations

import bisect
import os
from pathlib import Path

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.macho
import pwndbg.aglib.vmmap_custom
import pwndbg.dbg_mod
import pwndbg.lib.cache
import pwndbg.lib.memory
from pwndbg.dbg_mod import MemoryMap
from pwndbg.lib.arch import Platform
from pwndbg.lib.config import PARAM_BOOLEAN
from pwndbg.lib.memory import Page

pwndbg.config.add_param(
    "vmmap-prefer-relpaths",
    True,
    "show relative paths by default in vmmap",
    param_class=PARAM_BOOLEAN,
)


def _refine_memory_map(pages: MemoryMap) -> MemoryMap:
    if not (
        pwndbg.aglib.arch.platform == Platform.DARWIN
        and pwndbg.aglib.macho.shared_cache() is not None
    ):
        return pages

    # Darwin platforms use something called the Shared Cache for system
    # libraries. Debuggers may report mapping ranges that belong to the
    # shared cache in many ways, but we would like to tag those with a
    # little more information.
    final_pages = []

    shared_cache = pwndbg.aglib.macho.shared_cache()
    shared_cache_start = shared_cache.base
    shared_cache_end = shared_cache_start + shared_cache.size

    images = list(shared_cache.images_sorted)
    images_base = [image[1] for image in images]

    ptrsize: int = pwndbg.aglib.arch.ptrsize

    for page in pages.ranges():
        if page.end < shared_cache_start or page.start >= shared_cache_end:
            # No overlap with the shared cache.
            final_pages.append(page)
            continue

        # We do not support partial overlaps between other mappings and the
        # shared cache.
        #
        # While conceptually there's nothing stopping these from happening,
        # if we ever encounter such a situation, it likely means that we
        # either got something wrong, or that Darwin/LLDB has changed in
        # such a way that we are likely not able to gracefully handle.
        #
        assert page.start >= shared_cache_start and page.end <= shared_cache_end

        one_past_index = bisect.bisect_right(images_base, page.start)
        curr_base = page.start

        while True:
            if one_past_index > len(images):
                break

            if one_past_index == 0:
                # Indicates that this mapping is not part of any image, but
                # still part of the shared cache itself. Use a special name
                # for it.
                objfile = "[SharedCacheHeader]"
            elif images_base[one_past_index - 1] >= page.end:
                break
            else:
                # Name this mapping after the image it belongs to.
                objfile = images[one_past_index - 1][0].decode("ascii")
                curr_base = max(images_base[one_past_index - 1], page.start)

            if one_past_index == len(images):
                end = page.end
            else:
                end = min(page.end, images_base[one_past_index])

            final_pages.append(
                Page(
                    curr_base,
                    end - curr_base,
                    page.flags,
                    curr_base - shared_cache_start,
                    ptrsize,
                    objfile,
                    in_darwin_shared_cache=True,
                )
            )

            one_past_index += 1

    return type(pages)(final_pages)


@pwndbg.lib.cache.cache_until("start", "stop")
def get_memory_map() -> MemoryMap:
    return _refine_memory_map(pwndbg.dbg.selected_inferior().vmmap())


@pwndbg.lib.cache.cache_until("start", "stop")
def get() -> tuple[pwndbg.lib.memory.Page, ...]:
    return tuple(get_memory_map().ranges())


@pwndbg.lib.cache.cache_until("start", "stop")
def find(address: int | pwndbg.dbg_mod.Value | None) -> pwndbg.lib.memory.Page | None:
    if address is None:
        return None

    address = int(address)
    if address < 0:
        return None

    page = get_memory_map().lookup_page(address)

    if page is not None:
        return page

    return pwndbg.aglib.vmmap_custom.explore(address)


def addr_region_start(address: int | pwndbg.dbg_mod.Value) -> int | None:
    """
    Let's define a "region" as contiguous memory compromised of memory mappings
    which all have the same object file name. Also referred to as "File (Base)" by
    `xinfo`.

    Returns:
        The start of the memory region this address belongs to, or None if the address
        is not mapped.
    """
    address = int(address)
    if address < 0:
        return None

    mappings = sorted(get(), key=lambda p: p.vaddr)
    idx = -1
    for i in range(len(mappings)):
        if mappings[i].start <= address < mappings[i].end:
            idx = i
            break

    if idx == -1:
        # Maybe we can find the page by exploring.
        explored_page = pwndbg.aglib.vmmap_custom.explore(address)
        if not explored_page:
            return None

        # We know vmmap_custom.explore() can only find one page, it does
        # not cascade a whole region so there is no need to look backwards.
        return explored_page.start

    # Look backwards from i to find all the mappings with the same name.
    objname = mappings[idx].objfile
    while idx > 0 and objname == mappings[idx - 1].objfile:
        idx -= 1

    # There might be other mappings with the name "objname" in the address space
    # but they are not contiguous with us, so we don't care.
    return mappings[idx].start


def named_region_start(mapping_name: str, exact_match: bool = True) -> int | None:
    """
    Returns the lowest address which is mapped with `mapping_name`.

    This works both for object file names and stuff like "[heap]", but note that not
    all mappings with the same name are necessarily contiguous (especially if they
    aren't backed by an object file).

    Will not invoke vmmap_explore.

    If exact_match is True looks for exact path match, otherwise will match
    the os.path.basename()s.
    """
    mappings = sorted(get(), key=lambda p: p.vaddr)

    if exact_match:
        for mapping in mappings:
            # Resolve relative files and symlinks even for exact matches.
            # FIXME: This is a workaround for #3641 . Don't use Path()
            # after that is fixed.
            if Path(mapping.objfile).resolve() == Path(mapping_name).resolve():
                return mapping.start

        return None
    # Note that os.path.basename("[heap]") == "[heap]".
    mapping_basename = os.path.basename(mapping_name)
    for mapping in mappings:
        if os.path.basename(mapping.objfile) == mapping_basename:
            return mapping.start

    return None



"""
Add, load, show, edit, or delete custom structures.

For the compilation of the structures zig is being used under the hood, unless
`gcc-config-path` is specified.

You are advised to configure the 'cymbol-editor' config parameter to the path of your
favorite text editor. Otherwise cymbol expands $EDITOR and $VISUAL environment variables
to find the path to the default text editor.
"""

from __future__ import annotations

import argparse
import os
import subprocess
from pathlib import Path
from typing import TypeVar

from typing_extensions import ParamSpec

import pwndbg
import pwndbg.aglib.structures
import pwndbg.commands
import pwndbg.commands.context
import pwndbg.lib.config
from pwndbg.color import message
from pwndbg.commands import CommandCategory
from pwndbg.lib import Status

P = ParamSpec("P")
T = TypeVar("T")

cymbol_editor = pwndbg.config.add_param(
    "cymbol-editor",
    "",
    "path to the editor for editing custom structures",
    param_class=pwndbg.lib.config.PARAM_OPTIONAL_FILENAME,
)


def get_editor() -> str:
    if str(cymbol_editor) == "":
        # Lookup an editor to use for editing the a structure.
        editor: str | None = os.getenv("EDITOR")
        if editor is None:
            editor = os.getenv("VISUAL")
        if editor is None:
            editor = "vi"
    else:
        # Use the specified editor.
        editor = str(cymbol_editor)
    return editor


def run_editor_on_file(filepath: Path) -> bool:
    try:
        subprocess.run(
            [get_editor(), str(filepath)],
            check=True,
        )
    except Exception:
        print(message.error("An error occurred during opening the source file."))
        print(message.error(f"Path to the structure: {filepath}"))
        print(message.error("Please try to manually edit the structure."))
        print(
            message.hint(
                '\nTry to set a path to an editor with:\n\tset "cymbol-editor" /usr/bin/nano'
            )
        )
        return False

    input(message.notice("Press enter when finished."))
    return True


def _edit_and_load(name: str, struct_path: Path, preamble: str = "") -> None:
    if not run_editor_on_file(struct_path):
        return

    # Check that the user actually input something.
    with open(struct_path) as f:
        data: str = f.read().strip()
        if data == preamble.strip() or data == "":
            print(message.warn("Empty file, skipping..."))
            return

    print(message.success("Saved"), end="")

    err: Status = pwndbg.aglib.structures.load_with_path(name, struct_path)
    if err.is_success():
        print(message.success(" and loaded!"))
    else:
        print(message.error(" but failed loading."))
        print(message.error(err.message))


def add(name: str, force: bool) -> None:
    struct_path = pwndbg.aglib.structures.get_struct_path(name)
    if struct_path.exists() and not force:
        option = input(
            message.notice(
                "A custom structure was found with the given name, would you like to overwrite it? [y/N] "
            )
        )
        if option.lower() != "y":
            print(message.notice("Aborted by user."))
            return

    preamble = "// Enter your structure in a C header style.\n"
    preamble += f"// Refer to this structure file as '{name}'.\n"

    with open(struct_path, "w") as f:
        f.write(preamble)

    _edit_and_load(name, struct_path, preamble)


def add_from_header(header_file: str, name: str | None, force: bool, quiet: bool) -> None:
    if name is None:
        name = os.path.splitext(os.path.basename(header_file))[0]
    name = name.strip()

    if name == "":
        print(message.error("Invalid structure name provided or generated."))
        return

    struct_path: Path = pwndbg.aglib.structures.get_struct_path(name)
    if struct_path.exists():
        if not force:
            option = input(message.notice(f"Structure '{name}' already exists. Overwrite? [y/N] "))
            if option.lower() != "y":
                print(message.notice("Aborted by user."))
                return

    try:
        with open(header_file) as src, open(struct_path, "w") as f:
            content = src.read().strip()
            if not content:
                print(message.notice("Header file is empty, skipping..."))
                return
            f.write(content)
            if not quiet:
                print(message.success("Saved"), end="")
    except OSError as e:
        print(message.error(f"Failed to process header file: {e}"))
        return

    err: Status = pwndbg.aglib.structures.load_with_path(name, struct_path)
    if err.is_success():
        if not quiet:
            print(message.success(" and loaded!"))
    else:
        print(".\n" + err.message)


def load(name: str) -> None:
    struct_path: Path | None = pwndbg.aglib.structures.get_struct_path_if_exists(name)
    if struct_path is None:
        print(message.error("No custom structure was found with the given name!"))
        return

    err = pwndbg.aglib.structures.load_with_path(name, struct_path)
    if err.is_failure():
        print(message.error(err.message))
    else:
        print(message.success(f"Loaded custom structs! (from {struct_path})"))


def edit(name: str) -> None:
    struct_path: Path | None = pwndbg.aglib.structures.get_struct_path_if_exists(name)
    if struct_path is None:
        print(message.error("No custom structure was found with the given name!"))
        return

    _edit_and_load(name, struct_path)


def remove(name: str) -> None:
    err: Status = pwndbg.aglib.structures.remove(name)
    if err.is_success():
        print(message.success("Structs are removed!"))
    else:
        print(message.error(err.message))


def show(name: str) -> None:
    struct_path: Path | None = pwndbg.aglib.structures.get_struct_path_if_exists(name)
    if struct_path is None:
        print(message.error("No custom structure was found with the given name!"))
        return

    # Call non-caching version of the function
    highlighted_source = pwndbg.commands.context.get_highlight_source_uncached(str(struct_path))
    print("\n".join(highlighted_source))


def show_all() -> None:
    print(message.notice("Available custom structure names:\n"))
    names: list[str] = pwndbg.aglib.structures.saved_names()
    for name in names:
        if not name.startswith("_internal_"):
            print(f"  - {name}")
    for name in names:
        if name.startswith("_internal_"):
            print(f"  - {name}")


parser = argparse.ArgumentParser(
    description="""
Add custom C structures to the debugger.

Unless you specify `gcc-compiler-path`, zig is used under to hood to compile the C files to
whichever target architecture you are currently debugging.
"""
)

subparsers = parser.add_subparsers(dest="subcommand", help="Available subcommands")
subparsers.required = True

add_parser = subparsers.add_parser(
    "add",
    help="Add a custom structure and load it into the debugger",
    description="Add a custom structure and load it into the debugger.",
)
add_parser.add_argument("name", help="Name of custom structure", type=str)
add_parser.add_argument(
    "--force", action="store_true", help="Overwrite if structure already exists"
)

remove_parser = subparsers.add_parser(
    "remove",
    help="Remove a custom structure and unload it from the debugger",
    description="Remove a custom structure and unload it from the debugger.",
)
remove_parser.add_argument("name", help="Name of custom structure", type=str)

edit_parser = subparsers.add_parser(
    "edit",
    help="Edit a custom structure and reload it",
    description="Edit a custom structure and reload it.",
)
edit_parser.add_argument("name", help="Name of custom structure", type=str)

load_parser = subparsers.add_parser(
    "load",
    help="Load a previously added structure into the debugger",
    description="Load a previously added custom structure into the debugger.",
)
load_parser.add_argument("name", help="Name of custom structure", type=str)

show_parser = subparsers.add_parser(
    "show", help="Show a structure's definition", description="Show a structure's definition."
)
show_parser.add_argument("name", help="Name of custom structure", type=str)

file_parser = subparsers.add_parser(
    "file",
    help="Add and load a structure from a header file",
    description="Add and load a structure from a header file.",
)
file_parser.add_argument("path", help="Path to header file", type=str)
file_parser.add_argument("--name", help="Optional structure name", type=str)
file_parser.add_argument("--force", action="store_true", help="Overwrite if exists")
file_parser.add_argument(
    "--quiet", action="store_true", help="Do not output any message on success"
)

show_all_parser = subparsers.add_parser(
    "show-all",
    help="Show the names of all stored structures",
    description="Show the names of all stored (i.e. previusly added) structures.",
)


@pwndbg.commands.Command(
    parser,
    category=CommandCategory.MISC,
    notes=f"""
If a loaded structure defines a type that already exists, the debugger may prefer the
original type or behave unexpectedly. It’s recommended to use unique struct names to avoid
type conflicts.

Added structures are saved in {pwndbg.aglib.structures.storage_location}/.
""",
    examples="""
> cymbol file --force --quiet ./structs.h
Having something like this in your folder-local `.gdbinit` can be handy.
""",
)
def cymbol(
    subcommand: str,
    name: str = "",
    path: str = "",
    force: bool = False,
    quiet: bool = False,
) -> None:
    match subcommand:
        case "add":
            add(name, force)
        case "remove":
            remove(name)
        case "edit":
            edit(name)
        case "load":
            load(name)
        case "file":
            add_from_header(path, name, force, quiet)
        case "show":
            assert name is not None
            show(name)
        case "show-all":
            show_all()



from __future__ import annotations

import argparse

import pwndbg.color.message as message
import pwndbg.commands
from pwndbg.commands import CommandCategory
from pwndbg.lib.common import hex2ptr_common

# Define an argument parser for the command
parser = argparse.ArgumentParser(
    description="Converts a space-separated hex string to a little-endian address.",
)
parser.add_argument(
    "hex_string", type=str, help="Hexadecimal string to convert (e.g., '00 70 75 c1 cd ef 59 00')."
)


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
def hex2ptr(hex_string) -> None:
    hex_string = hex_string.replace(" ", "")
    try:
        pointer = hex2ptr_common(hex_string)
        print(message.success(f"{hex(pointer)}"))
    except Exception as e:
        print(message.error(str(e)))



"""
Dumps all pwndbg-specific configuration points.
"""

from __future__ import annotations

import argparse

import pwndbg
import pwndbg.commands
import pwndbg.lib.config
from pwndbg.color import generateColorFunction
from pwndbg.color import ljust_colored
from pwndbg.color import strip
from pwndbg.color.message import hint
from pwndbg.commands import CommandCategory
from pwndbg.lib.config import Scope

if pwndbg.dbg.is_gdblib_available():
    import pwndbg.gdblib.config


def print_row(
    name: str,
    value: str,
    default: str,
    set_show_doc: str,
    ljust_optname: int,
    ljust_doc: int,
    empty_space: int = 2,
):
    name = ljust_colored(name, ljust_optname + empty_space)
    set_show_doc = ljust_colored(set_show_doc, ljust_doc + empty_space)
    defval = extend_value_with_default(value, default)
    result = f"{name} {set_show_doc} {defval} "
    print(result)
    return result


def extend_value_with_default(value, default):
    if strip(value) != strip(default):
        return f"{value} ({default})"
    return value


def get_config_parameters(scope: Scope, filter_pattern: str):
    values = [
        v
        for k, v in pwndbg.config.params.items()
        if isinstance(v, pwndbg.lib.config.Parameter) and v.scope == scope
    ]

    if filter_pattern:
        filter_pattern = filter_pattern.lower()
        values = [
            v
            for v in values
            if filter_pattern in v.name.lower() or filter_pattern in v.set_show_doc.lower()
        ]

    return values


parser = argparse.ArgumentParser(description="Shows Pwndbg-specific configuration.")
parser.add_argument(
    "filter_pattern",
    type=str,
    nargs="?",
    default=None,
    help="Filter to apply to config parameters names/descriptions",
)


def display_config(
    filter_pattern: str, scope: Scope, has_file_command: bool = True, show_hints: bool = True
) -> None:
    values = get_config_parameters(scope, filter_pattern)

    if not values:
        print(hint(f'No {scope.name} parameter found with filter "{filter_pattern}"'))
        return

    longest_optname = max(map(len, (v.name for v in values)))
    longest_doc = max(map(len, (v.set_show_doc for v in values)))

    header = print_row("Name", "Value", "Default", "Documentation", longest_optname, longest_doc)
    print("-" * len(header))

    for v in sorted(values):
        if isinstance(v, pwndbg.color.theme.ColorParameter):
            # Only the theme scope should use ColorParameter
            assert scope == Scope.theme

            value = generateColorFunction(v.value)(v.value)
            default = generateColorFunction(v.default)(v.default)
        else:
            value = v.pretty()
            default = v.pretty_default()

        print_row(v.name, value, default, v.set_show_doc, longest_optname, longest_doc)

    if show_hints:
        print(
            hint(
                f"You can set a {scope.name} variable with `set <{scope.name}-var> <value>`, and read more about it with `help set <{scope.name}-var>`."
            )
        )
        if has_file_command:
            print(
                hint(
                    f"You can generate a configuration file using `{scope.name}file` "
                    "- then put it in your .gdbinit after initializing pwndbg."
                )
            )


@pwndbg.commands.Command(parser, category=CommandCategory.PWNDBG)
def config(filter_pattern) -> None:
    display_config(filter_pattern, Scope.config)


configfile_parser = argparse.ArgumentParser(
    description="Generates a configuration file for the current Pwndbg options."
)
configfile_parser.add_argument(
    "--show-all", action="store_true", help="Display all configuration options."
)

parser = argparse.ArgumentParser(description="Shows Pwndbg-specific theme configuration.")
parser.add_argument(
    "filter_pattern",
    type=str,
    nargs="?",
    default=None,
    help="Filter to apply to theme parameters names/descriptions",
)


@pwndbg.commands.Command(parser, category=CommandCategory.PWNDBG)
def theme(filter_pattern) -> None:
    display_config(filter_pattern, Scope.theme)


if pwndbg.dbg.is_gdblib_available():
    # Register the configfile command
    @pwndbg.commands.Command(configfile_parser, category=CommandCategory.PWNDBG)
    def configfile(show_all=False) -> None:
        configfile_print_scope(Scope.config, show_all)


themefile_parser = argparse.ArgumentParser(
    description="Generates a configuration file for the current Pwndbg theme options."
)
themefile_parser.add_argument(
    "--show-all", action="store_true", help="Force displaying of all theme options."
)


if pwndbg.dbg.is_gdblib_available():
    # Register the themefile command.
    @pwndbg.commands.Command(themefile_parser, category=CommandCategory.PWNDBG)
    def themefile(show_all=False) -> None:
        configfile_print_scope(Scope.theme, show_all)


parser = argparse.ArgumentParser(description="Shows heap related configuration.")
parser.add_argument(
    "filter_pattern",
    type=str,
    nargs="?",
    default=None,
    help="Filter to apply to config parameters names/descriptions",
)


@pwndbg.commands.Command(parser, category=CommandCategory.PWNDBG)
def heap_config(filter_pattern: str) -> None:
    display_config(filter_pattern, Scope.heap, has_file_command=False)
    print(
        hint(
            "Some parameters (e.g. main-arena) will be used only when resolve-heap-via-heuristic is `auto` or `force`"
        )
    )


def configfile_print_scope(scope: Scope, show_all: bool = False) -> None:
    params = pwndbg.config.get_params(scope)

    if not show_all:
        params = list(filter(lambda p: p.is_changed, params))

    if params:
        if not show_all:
            print(hint("Showing only changed values:"))
        for p in params:
            native_default = pwndbg.gdblib.config_mod.Parameter._value_to_gdb_native(
                p.default, gdb_param_class=pwndbg.gdblib.config.CLASS_MAPPING[p.param_class]
            )
            native_value = pwndbg.gdblib.config_mod.Parameter._value_to_gdb_native(
                p.value, gdb_param_class=pwndbg.gdblib.config.CLASS_MAPPING[p.param_class]
            )
            print(f"# {p.name}: {p.set_show_doc}")
            print(f"# default: {native_default}")
            print(f"set {p.name} {native_value}")
            print()
    else:
        # FIXME: Message would be wrong for a heapfile command.
        print(hint(f"No changed values. To see current values use `{scope.name}`."))



from __future__ import annotations

import pwndbg.aglib.symbol


def version_parse(data: bytearray) -> tuple[int, ...]:
    """
    Parse the version bytestring as read from memory into
    an integer tuple.
    """
    # Example from compiled musl: 1.2.5-git-103-g7bcf8783
    data = data.split(b"-")[0]
    return tuple(int(part) for part in data.split(b"."))


def has_exported_symbols(mapping_name: str) -> bool:
    # fscanf must be implemented by a libc.
    return pwndbg.aglib.symbol.lookup_symbol("fscanf", objfile_endswith=mapping_name) is not None



from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import pwndbg
import pwndbg.color
import pwndbg.color.memory
import pwndbg.lib.config
from pwndbg.color import theme

max_decimal_number = pwndbg.config.add_param(
    "max-decimal-number",
    9,
    "show all numbers greater than this in hex",
    param_class=pwndbg.lib.config.PARAM_ZUINTEGER_UNLIMITED,
    help_docstring="""
For negative numbers, their absolute value is used.

Set the parameter to 'unlimited' if you want all values in decimal.
Specially, set the parameter to zero if you want all values in hex.

The assembly instruction operands come from capstone, and are thus
not controlled by this setting. For consistency with them, leave
this setting at 9 (the default).
""",
    # We could look into also overwriting the capstone operands string, similarly
    # to what is done here: https://github.com/pwndbg/pwndbg/blob/26db4533aa08d77c4bbc359b4760a0944e0c6b23/pwndbg/aglib/disasm/arch.py#L322-L331
)


def int_to_string(num: int) -> str:
    """
    Converts an integer value to string.

    Decides whether to format it in decimal or
    hex depending on the max-decimal-number config.
    """
    if max_decimal_number == -1:
        return f"{num}"
    if max_decimal_number == 0:
        return f"{num:#x}"
    if abs(num) > max_decimal_number:
        return f"{num:#x}"
    return f"{num}"


def int_pair_to_string(num1: int, num2: int) -> tuple[str, str]:
    """
    Converts an integer pair to a string pair.

    Decides whether to format them in decimal or
    hex depending on the max-decimal-number config.

    If either value should be hex, both are hex.
    """
    if max_decimal_number == -1:
        return f"{num1}", f"{num2}"
    if max_decimal_number == 0:
        return f"{num1:#x}", f"{num2:#x}"
    if abs(num1) > max_decimal_number or abs(num2) > max_decimal_number:
        return f"{num1:#x}", f"{num2:#x}"
    return f"{num1}", f"{num2}"


config_property_name_color = theme.add_color_param(
    "prop-name-color",
    "bold",
    "color used to highlight the name in name-value pairs",
    help_docstring="""
Used heavily in mallocng commands.
""",
)

config_property_value_color = theme.add_color_param(
    "prop-value-color",
    "yellow",
    "color used to highlight the value in name-value pairs",
    help_docstring="""
Used heavily in mallocng commands.
""",
)

config_property_title_color = theme.add_color_param(
    "prop-title-color",
    "green",
    "color used to highlight the title of name-value pair groups",
    help_docstring="""
Used heavily in mallocng commands.
""",
)


@dataclass
class Property:
    """
    A (property name, property value) pair
    with optional extra information.

    Used by from_properties().
    """

    name: str
    value: Any
    # Alternate value, will be shown in brackets e.g.
    #   slack:   0x2 (0x20)
    alt_value: Any = None
    # Extra explanation, may be list, e.g.
    #   hdr reserved: 0x5  describes: end - p - n
    #                      use ftr reserved
    extra: str | list[str] = ""
    # Will print the value as hex and use the address's
    # mapping's color.
    is_addr: bool = False
    # Will turn an integer into its hex representation.
    use_hex: bool = True
    # Override the color used by from_properties().
    name_color_func: Callable[[str], str] | None = None
    value_color_func: Callable[[str], str] | None = None


def from_properties(
    title: str,
    properties: list[Property],
    *,
    preamble: str = "",
    value_offset: int = 14,
    extra_offset: int = 16,
    title_color_func: Callable[[str], str] | None = None,
    name_color_func: Callable[[str], str] | None = None,
    value_color_func: Callable[[str], str] | None = None,
    indent_size: int = 2,
) -> str:
    """
    When you have (property name, property value) pairs
    that you want to print, each on a new line.

    A common usecase is printing a struct.

    Example:
        general
          start:          0x7ffff7ff6040
          user start:     0x7ffff7ff6040    aka `p`
          end:            0x7ffff7ff606c    start + stride - 4
          stride:         0x30              distance between adjacent slots
          user size:      0x20              aka "nominal size", `n`
          slack:          0x0 (0x0)         slot's unused memory / 0x10

    Arguments:
        title: The title of this property group. An empty string may be provided for a
            titleless group.
        properties: The list of properties to format.
        preamble: A string that will be printed between the title and the properties,
            may be used to denote the address of an object like e.g. `@ 0x408000 - 0x408fe0`
        value_offset: The number of characters from the start of the name of a property to the
            start of its value.
        extra_offset: The number of characters from the start of the value of a property to the
            start of its extra text.
        title_color_func: The function to use to color the title.
        name_color_func: The function to use to color names.
        value_color_func: The function to use to color values. This function isn't applied to
            is_addr=True properties.
        indent_size: The indentation to use i.e. the offset from the title to the names.
    """

    if name_color_func is None:
        name_color_func = config_property_name_color.color_function

    if value_color_func is None:
        value_color_func = config_property_value_color.color_function

    if title_color_func is None:
        title_color_func = config_property_title_color.color_function

    text = ""

    if title:
        text += title_color_func(title) + "\n"

    if preamble:
        text += " " * indent_size
        text += preamble + "\n"

    # Transform prop values to string representation
    for prop in properties:
        if isinstance(prop.value, int):
            prop.value = f"{prop.value:#x}" if prop.use_hex else f"{prop.value}"

        if isinstance(prop.alt_value, int):
            prop.alt_value = f"{prop.alt_value:#x}" if prop.use_hex else f"{prop.alt_value}"

    indentation_str = indent_size * " "
    extra_list_pad_str = indentation_str + value_offset * " " + "  " + extra_offset * " "

    for prop in properties:
        # The property may override the generic color functions.
        prop_name_cfunc = (
            prop.name_color_func if prop.name_color_func is not None else name_color_func
        )
        prop_value_cfunc = (
            prop.value_color_func if prop.value_color_func is not None else value_color_func
        )

        text += (
            indentation_str
            + pwndbg.color.ljust_colored(prop_name_cfunc(prop.name) + ":", value_offset)
            + "  "
        )

        if prop.is_addr:
            base = 16 if prop.use_hex else 10
            colored_val = pwndbg.color.memory.get(int(prop.value, base))
        else:
            colored_val = prop_value_cfunc(prop.value)

        colored_alt_val = ""
        if prop.alt_value is not None:
            colored_alt_val = f" ({prop_value_cfunc(prop.alt_value)})"

        text += pwndbg.color.ljust_colored(colored_val + colored_alt_val, extra_offset)

        if isinstance(prop.extra, str):
            text += "  " + prop.extra
        else:
            # list of strings, we want each one under the other
            assert isinstance(prop.extra, list)

            text += "  " + prop.extra[0]
            for i in range(1, len(prop.extra)):
                text += "\n"
                text += extra_list_pad_str
                text += "  " + prop.extra[i]

        text += "\n"

    return text


def nlines_to_range(nlines: int, current: int, total: int) -> tuple[int, int]:
    """
    When you want to get nlines of output around a certain interesting line, returns
    the range to use.

    `end - start` will not be `== nlines` only if `nlines > total`.

    The `current` line may not necessarily be centered in the range, if decentering
    it gets `end - start` closer to `nlines`.

    Arguments:
        nlines: The exact amount of lines you want.
        current: The index of the interesting line (e.g. where PC is in the source context)
        total: How many lines total you actually have.

    Returns:
        A tuple giving the range of indecies to use. The format is [start, end).
    """
    if nlines > total:
        return (0, total)

    # Note that in both calculations, ideal_end is exclusive (so we have +1)
    if nlines % 2 == 1:
        ideal_start: int = current - (nlines // 2)
        ideal_end: int = current + (nlines // 2) + 1
    else:
        # Since it is impossible to center exactly due to parity, we will make
        # `current` have the lower index because this is usually more visually pleasing.
        ideal_start = current - (nlines // 2) + 1
        ideal_end = current + (nlines // 2) + 1

    # Now it may be that we are outside of the allowed range, but if we are, we
    # are only outside on one side because we already checked `nlines > total`.
    if ideal_start < 0:
        # Now (-ideal_start) is the amount of lines we have to steal from the end
        # of the range.
        start = 0
        # ideal_end + (-ideal_start) = ideal_end - ideal_start
        end = ideal_end - ideal_start
        # We don't need to do `end = min(end, total)` because that would imply
        # that `nlines > total`.
    elif ideal_end > total:
        # Now (ideal_end - total) is the amount of lines we have to steal from the start
        # of the range.
        # ideal_start - (ideal_end - total) = ideal_start - ideal_end + total
        start = ideal_start - ideal_end + total
        end = total
        # We don't need to do `start = max(start, 0)` because that would imply
        # that `nlines > total`.
    else:
        start = ideal_start
        end = ideal_end

    return (start, end)


def format_source(source: list[str], nlines: int, interesting_line: int) -> list[str]:
    """
    Format source code.

    Use correct tab size, add the code prefix (►), add line numbers, align
    properly.

    Arguments:
        source: Already highlighted source code. List of lines.
        nlines: The amount of lines we want back.
        interesting_line: The line around which to center the output.
    """
    start, end = nlines_to_range(nlines, interesting_line, len(source))
    num_width = len(str(end))

    # split the code
    source = source[start:end]

    # Compute the prefix_sign length
    prefix_sign = pwndbg.color.context.prefix(str(pwndbg.config.code_prefix))
    prefix_width = len(prefix_sign)

    # Format the output
    formatted_source = []
    for line_number, code in enumerate(source, start=start + 1):
        # Honor the tab-size setting.
        if pwndbg.config.context_code_tabstop > 0:
            code = code.replace("\t", " " * int(pwndbg.config.context_code_tabstop))

        # Remove extra whitespace
        code = code.rstrip()

        fmt = " {prefix_sign:{prefix_width}} {line_number:>{num_width}} {code}"

        if pwndbg.config.highlight_source and line_number == interesting_line:
            fmt = pwndbg.color.context.highlight(fmt)

        line = fmt.format(
            prefix_sign=prefix_sign if line_number == interesting_line else "",
            prefix_width=prefix_width,
            line_number=line_number,
            num_width=num_width,
            code=code,
        )
        formatted_source.append(line)

    return formatted_source



from __future__ import annotations

import importlib
import sys

import gdb

import pwndbg
import pwndbg.commands
import pwndbg.gdblib.events
import pwndbg.lib.cache
from pwndbg.commands import CommandCategory


def rreload(module, _exclude_mods=None) -> None:
    """Recursively reload modules.
    Impl based on https://stackoverflow.com/a/66661311/1508881"""
    for module in list(sys.modules.keys()):
        if "pwndbg" in module:
            del sys.modules[module]

    # Mark that we are reloading; this is used to prevent the Command decorator from
    # erroring out on re-registering the same commands we had registered before
    gdb.pwndbg_is_reloading = True
    importlib.import_module("pwndbg")


@pwndbg.commands.Command("Reload Pwndbg.", category=CommandCategory.PWNDBG)
def reload(*a) -> None:
    pwndbg.gdblib.events.on_reload()
    rreload(pwndbg)
    pwndbg.gdblib.events.after_reload()


@pwndbg.commands.Command("Makes Pwndbg reinitialize all state.", category=CommandCategory.PWNDBG)
def reinit_pwndbg() -> None:
    """
    Makes pwndbg reinitialize all state.
    """
    pwndbg.lib.cache.clear_caches()
    pwndbg.gdblib.events.after_reload()



from __future__ import annotations

import cProfile
import os
import sys
import time

import lldb

from pwndbginit.common import verify_venv


def main(debugger: lldb.SBDebugger, lldb_version: tuple[int, ...], debug: bool = False) -> None:
    if "pwndbg" in sys.modules:
        print("Detected double-loading of Pwndbg.")
        print("This should not happen. Please report this issue if you're not sure how to fix it.")
        sys.exit(1)

    verify_venv()
    profiler = cProfile.Profile()

    start_time = None
    if os.environ.get("PWNDBG_PROFILE") == "1":
        start_time = time.time()
        profiler.enable()

    import pwndbg  # noqa: F811
    import pwndbg.dbg_mod.lldb

    pwndbg.dbg_mod.lldb.LLDB_VERSION = lldb_version

    pwndbg.dbg = pwndbg.dbg_mod.lldb.LLDB()
    pwndbg.dbg.setup(debugger, "pwndbglldbhandler", debug=debug)

    import pwndbg.profiling

    pwndbg.profiling.init(profiler, start_time)
    if os.environ.get("PWNDBG_PROFILE") == "1":
        pwndbg.profiling.profiler.stop("pwndbg-load.pstats")
        pwndbg.profiling.profiler.start()



"""
Prints out pointer chains starting at some address in memory.

Generally used to print out the stack or register values.
"""

from __future__ import annotations

import argparse
import collections
import math

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.aglib.vmmap
import pwndbg.chain
import pwndbg.color.telescope as T
import pwndbg.commands
from pwndbg.color import theme
from pwndbg.commands import CommandCategory

telescope_lines = pwndbg.config.add_param(
    "telescope-lines", 8, "number of lines to printed by the telescope command"
)
skip_repeating_values = pwndbg.config.add_param(
    "telescope-skip-repeating-val",
    True,
    "whether to skip repeating values of the telescope command",
)
skip_repeating_values_minimum = pwndbg.config.add_param(
    "telescope-skip-repeating-val-min",
    3,
    "minimum amount of repeated values before skipping lines",
)
print_framepointer_offset = pwndbg.config.add_param(
    "telescope-framepointer-offset",
    True,
    "print offset to framepointer for each address, if sufficiently small",
)
print_retaddr_in_frame = pwndbg.config.add_param(
    "telescope-frame-print-retaddr", True, "print one pointer past the stack frame"
)
dont_skip_registers = pwndbg.config.add_param(
    "telescope-dont-skip-registers", True, "don't skip a repeated line if a registers points to it"
)

offset_separator = theme.add_param(
    "telescope-offset-separator", "│", "offset separator of the telescope command"
)
offset_delimiter = theme.add_param(
    "telescope-offset-delimiter", ":", "offset delimiter of the telescope command"
)
repeating_marker = theme.add_param(
    "telescope-repeating-marker", "... ↓", "repeating values marker of the telescope command"
)


parser = argparse.ArgumentParser(
    description="Recursively dereferences pointers starting at the specified address."
)
parser.add_argument(
    "-r",
    "--reverse",
    dest="reverse",
    action="store_true",
    default=False,
    help="Show <count> previous addresses instead of next ones",
)

parser.add_argument(
    "-f",
    "--frame",
    dest="frame",
    action="store_true",
    default=False,
    help="Show the stack frame, from rsp to rbp",
)

parser.add_argument(
    "-i",
    "--inverse",
    dest="inverse",
    action="store_true",
    default=False,
    help="Show the stack reverse growth",
)


parser.add_argument(
    "address", nargs="?", default="$sp", type=int, help="The address to telescope at."
)

parser.add_argument(
    "count", nargs="?", default=telescope_lines, type=int, help="The number of lines to show."
)


@pwndbg.commands.Command(parser, category=CommandCategory.MEMORY)
@pwndbg.commands.OnlyWhenRunning
def telescope(
    address=None, count=telescope_lines, to_string=False, reverse=False, frame=False, inverse=False
):
    """
    Recursively dereferences pointers starting at the specified address
    ($sp by default)
    """
    ptrsize = pwndbg.aglib.typeinfo.ptrsize
    if telescope.repeat:
        address = telescope.last_address + ptrsize
        telescope.offset += 1
    else:
        telescope.offset = 0

    address = address if address else pwndbg.aglib.regs.sp
    if address is None:
        print("Cannot display stack frame because stack pointer is unavailable")
        return

    address = int(address) & pwndbg.aglib.arch.ptrmask
    input_address = address
    count = max(int(count), 1) & pwndbg.aglib.arch.ptrmask
    delimiter = T.delimiter(offset_delimiter)
    separator = T.separator(offset_separator)

    # Allow invocation of "telescope 20" to dump 20 bytes at the stack pointer
    if address < pwndbg.aglib.memory.MMAP_MIN_ADDR and not pwndbg.aglib.memory.peek(address):
        count = address
        address = pwndbg.aglib.regs.sp

    # Allow invocation of telescope -r to dump previous addresses
    if reverse:
        address -= (count - 1) * ptrsize

    # Allow invocation of telescope -f (--frame) to dump all addresses in a frame
    if frame:
        if not pwndbg.aglib.regs.frame:
            print("The frame register is not defined for this architecture.")
            return
        sp = pwndbg.aglib.regs.sp
        bp = pwndbg.aglib.regs.read_reg(pwndbg.aglib.regs.frame)
        if sp > bp:
            print("Cannot display stack frame because base pointer is below stack pointer")
            return

        for page in pwndbg.aglib.vmmap.get():
            if sp in page and bp not in page:
                print(
                    "Cannot display stack frame because base pointer is not on the same page with stack pointer"
                )
                return

        address = sp
        count = int((bp - sp) / ptrsize) + 1
        count += 1 if print_retaddr_in_frame else 0

    # Allow invocation of "telescope a b" to dump all bytes from A to B
    if int(address) <= int(count):
        # adjust count if it is an address. use ceil division as count is number of
        # ptrsize values and we don't want to strip out a value if dest is unaligned
        count -= address
        count = max(math.ceil(count / ptrsize), 1)

    # Map of address to register string
    reg_values: collections.defaultdict[int, list[str]] = collections.defaultdict(list)
    for reg in pwndbg.aglib.regs.common:
        reg_values[pwndbg.aglib.regs.read_reg(reg)].append(reg)

    if not inverse:
        start = address
        stop = address + (count * ptrsize)
        step = ptrsize
    else:
        start = address + ((count - 1) * ptrsize)
        stop = address - ptrsize
        step = -1 * ptrsize

    # Find all registers which show up in the trace, map address to regs
    regs: dict[int, str] = {}
    for i in range(start, stop, step):
        values = list(reg_values[i])

        # Find all regs that point to somewhere in the current ptrsize step
        for width in range(1, pwndbg.aglib.arch.ptrsize):
            values.extend(f"{r}-{width}" for r in reg_values[i + width])

        regs[i] = " ".join(values)

    # Find the longest set of register information (length of string), used for padding
    if regs:
        longest_regs = max(map(len, regs.values()))
    else:
        longest_regs = 0

    # Print everything out
    result = []
    last = None
    collapse_buffer: list[str] = []
    skipped_padding = (
        2
        + len(offset_delimiter)
        + 4
        + len(offset_separator)
        + 1
        + longest_regs
        + 1
        - len(repeating_marker)
    )

    # Collapse repeating values exceeding minimum delta.
    def collapse_repeating_values() -> None:
        # The first line was already printed, hence increment by 1
        if collapse_buffer and len(collapse_buffer) + 1 >= skip_repeating_values_minimum:
            result.append(
                T.repeating_marker(
                    f"{repeating_marker}{' ' * skipped_padding}{len(collapse_buffer)} skipped"
                )
            )
        else:
            result.extend(collapse_buffer)
        collapse_buffer.clear()

    bp = None
    if print_framepointer_offset and pwndbg.aglib.regs.frame is not None:
        bp = pwndbg.aglib.regs.read_reg(pwndbg.aglib.regs.frame)

    for i, addr in enumerate(range(start, stop, step)):
        if not pwndbg.aglib.memory.peek(addr):
            collapse_repeating_values()
            result.append(f"<Could not read memory at {addr:#x}>")
            break
        if inverse:
            line_offset = addr - (stop + ptrsize) + (telescope.offset * ptrsize)
            idx_offset = int((start - stop - ptrsize) / ptrsize) - (i + telescope.offset)
        else:
            line_offset = addr - start + (telescope.offset * ptrsize)
            idx_offset = i + telescope.offset
        line = T.offset(f"{idx_offset:02x}{delimiter}{line_offset:04x}{separator}") + " ".join(
            (
                regs_or_frame_offset(addr, bp, regs, longest_regs),
                pwndbg.chain.format(addr),
            )
        )

        # Buffer repeating values.
        if skip_repeating_values:
            value = pwndbg.aglib.memory.read_pointer_width(addr)
            if (
                last == value
                and addr != input_address
                and (not dont_skip_registers or not regs[addr])
            ):
                collapse_buffer.append(line)
                continue
            collapse_repeating_values()
            last = value

        result.append(line)

    collapse_repeating_values()
    telescope.offset += i
    telescope.last_address = addr

    if not to_string:
        print("\n".join(result))

    return result


def regs_or_frame_offset(addr: int, bp: int | None, regs: dict[int, str], longest_regs: int) -> str:
    # bp only set if print_framepointer_offset=True
    if bp is None or regs[addr] or not -0xFFF <= addr - bp <= 0xFFF:
        # We do .rjust(3) because some arches have two-letter registers.
        return " " + T.register(regs[addr].ljust(longest_regs).rjust(3))
    # If offset to frame pointer as hex fits in hex 3 digits, print it
    return ("%+04x" % (addr - bp)).ljust(longest_regs + 1)


parser = argparse.ArgumentParser(
    description="Dereferences on stack data with specified count and offset."
)
parser.add_argument(
    "-f",
    "--frame",
    dest="frame",
    action="store_true",
    default=False,
    help="Show the stack frame, from rsp to rbp",
)

parser.add_argument(
    "-i",
    "--inverse",
    dest="inverse",
    action="store_true",
    default=False,
    help="Show reverse stack growth",
)

parser.add_argument("count", nargs="?", default=8, type=int, help="number of element to dump")
parser.add_argument(
    "offset",
    nargs="?",
    default=0,
    type=int,
    help="Element offset from $sp (support negative offset)",
)


@pwndbg.commands.Command(parser, category=CommandCategory.STACK)
@pwndbg.commands.OnlyWhenRunning
def stack(count, offset, frame, inverse) -> None:
    ptrsize = pwndbg.aglib.typeinfo.ptrsize
    telescope.repeat = stack.repeat
    telescope(
        address=pwndbg.aglib.regs.sp + offset * ptrsize, count=count, frame=frame, inverse=inverse
    )


parser = argparse.ArgumentParser(
    description="Dereferences on stack data, printing the entire stack frame with specified count and offset ."
)
parser.add_argument("count", nargs="?", default=8, type=int, help="number of element to dump")
parser.add_argument(
    "offset",
    nargs="?",
    default=0,
    type=int,
    help="Element offset from $sp (support negative offset)",
)


@pwndbg.commands.Command(parser, category=CommandCategory.STACK)
@pwndbg.commands.OnlyWhenRunning
def stackf(count, offset) -> None:
    ptrsize = pwndbg.aglib.typeinfo.ptrsize
    telescope.repeat = stack.repeat
    telescope(address=pwndbg.aglib.regs.sp + offset * ptrsize, count=count, frame=True)


telescope.last_address = 0
telescope.offset = 0



"""
Commands for setting temporary breakpoints on the next
instruction of some type (call, branch, etc.)
"""

from __future__ import annotations

import re
from itertools import chain

from capstone6pwndbg import CS_GRP_INT
from capstone6pwndbg import CS_GRP_RET

import pwndbg.aglib
import pwndbg.aglib.disasm.disassembly
import pwndbg.aglib.proc
import pwndbg.aglib.vmmap
import pwndbg.dbg_mod
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.color import message
from pwndbg.dbg_mod import BreakpointLocation

interrupts = {CS_GRP_INT}


def next_int(address=None, honor_current_branch=False):
    """
    If there is a syscall in the current basic black,
    return the instruction of the one closest to $PC.

    If honor_current_branch is True, then if the address is already a branch, return None.

    If no interrupt exists or a jump is in the way, return None.
    """
    if address is None:
        ins = pwndbg.aglib.disasm.disassembly.one(pwndbg.aglib.regs.pc)
        if not ins:
            return None
        if honor_current_branch and ins.jump_like:
            return None
        address = ins.next

    ins = pwndbg.aglib.disasm.disassembly.one(address)
    while ins:
        if ins.jump_like:
            return None
        if ins.groups & interrupts:
            return ins
        ins = pwndbg.aglib.disasm.disassembly.one(ins.next)

    return None


def next_branch(address=None, including_current=False) -> PwndbgInstruction | None:
    """
    Return the next branch instruction that the process will encounter with repeated usage of the "nexti" command.

    If including_current == True, then if the instruction at the address is already a branch, return it.
    """
    if address is None:
        ins = pwndbg.aglib.disasm.disassembly.one(pwndbg.aglib.regs.pc)
        if not ins:
            return None
        if including_current and ins.jump_like:
            return ins
        address = ins.next

    ins = pwndbg.aglib.disasm.disassembly.one(address)
    while ins:
        if ins.jump_like:
            return ins
        ins = pwndbg.aglib.disasm.disassembly.one(ins.next)

    return None


def next_matching_until_branch(address=None, mnemonic=None, op_str=None):
    """
    Finds the next instruction that matches the arguments between the given
    address and the branch closest to it.
    """
    if address is None:
        address = pwndbg.aglib.regs.pc

    ins = pwndbg.aglib.disasm.disassembly.one(address)
    while ins:
        # Check whether or not the mnemonic matches if it was specified
        mnemonic_match = ins.mnemonic.casefold() == mnemonic.casefold() if mnemonic else True

        # Check whether or not the operands match if they were specified
        op_str_match = True
        if op_str is not None:
            op_str_match = False

            # Remove whitespace and fold the case of both targets.
            ops = "".join(ins.op_str.split()).casefold()
            if isinstance(op_str, str):
                op_str = "".join(op_str.split()).casefold()
            elif isinstance(op_str, list):
                op_str = "".join(chain.from_iterable(op.split() for op in op_str)).casefold()
            else:
                raise ValueError("op_str value is of an unsupported type")
            op_str_match = ops == op_str

        # If all of the parameters that were specified match, this is the
        # instruction we want to stop at.
        if mnemonic_match and op_str_match:
            return ins

        if ins.jump_like:
            # No matching instruction until the next branch, and we're
            # not trying to match the branch instruction itself.
            return None

        ins = pwndbg.aglib.disasm.disassembly.one(ins.next)
    return None


async def break_next_branch(
    ec: pwndbg.dbg_mod.ExecutionController, address=None, including_current=False
):
    """
    If including_current == True, do not step in case we are currently on a branch
    """
    ins = next_branch(address, including_current=including_current)

    inf = pwndbg.dbg.selected_inferior()
    if ins:
        # If the branch we found was not at the current program counter, we should step to it.
        # Otherwise, return the current instruction.
        if ins.address != pwndbg.aglib.regs.pc:
            with inf.break_at(BreakpointLocation(ins.address), internal=True) as bp:
                await ec.cont(bp)
        return ins


async def break_next_interrupt(
    ec: pwndbg.dbg_mod.ExecutionController, address=None, honor_current_branch=False
) -> PwndbgInstruction | None:
    """
    Break at the next interrupt if there is one in the current basic block
    and no jumps are between the current instruction and the interrupt.

    If no such interrupt exists or a jump is in the way, return None.
    """
    ins = next_int(address, honor_current_branch=honor_current_branch)
    proc = pwndbg.dbg.selected_inferior()
    if ins:
        with proc.break_at(BreakpointLocation(ins.address), internal=True) as bp:
            await ec.cont(bp)
        return ins

    return None


async def break_next_call(ec: pwndbg.dbg_mod.ExecutionController, symbol_regex=None):
    symbol_regex = re.compile(symbol_regex) if symbol_regex else None

    while pwndbg.aglib.proc.alive():
        # Break on signal as it may be a segfault
        if pwndbg.aglib.proc.stopped_with_signal():
            return

        ins = await break_next_branch(ec)

        if not ins:
            break

        # continue if not a call
        if not ins.call_like:
            continue

        # return call if we:
        # 1) don't search for a symbol
        # 2) match target address
        # 3) match symbol name
        if (
            not symbol_regex
            or (ins.target_const and symbol_regex.match(hex(ins.target)))
            or (ins.target_string and symbol_regex.match(ins.target_string))
        ):
            return ins


async def break_next_ret(ec: pwndbg.dbg_mod.ExecutionController, address=None):
    while pwndbg.aglib.proc.alive():
        # Break on signal as it may be a segfault
        if pwndbg.aglib.proc.stopped_with_signal():
            return

        ins = await break_next_branch(ec, address)

        if not ins:
            break

        if CS_GRP_RET in ins.groups:
            return ins


async def break_on_next_matching_instruction(
    ec: pwndbg.dbg_mod.ExecutionController, mnemonic=None, op_str=None
) -> bool:
    """
    Breaks on next instuction that matches the arguments.
    """
    # Make sure we have something to break on.
    if mnemonic is None and op_str is None:
        return False

    proc = pwndbg.dbg.selected_inferior()
    while pwndbg.aglib.proc.alive():
        ins = next_matching_until_branch(mnemonic=mnemonic, op_str=op_str)
        if ins is not None:
            if ins.address != pwndbg.aglib.regs.pc:
                # Only set breakpoints at a different PC location, otherwise we
                # will continue until we hit a breakpoint that's not related to
                # this opeeration, or the program halts.
                with proc.break_at(BreakpointLocation(ins.address), internal=True) as bp:
                    await ec.cont(bp)
                return ins
            # We don't want to be spinning in place, nudge execution forward
            # and try again.
            pass
        else:
            # Move to the next branch instruction.
            nb = next_branch(pwndbg.aglib.regs.pc, including_current=True)
            if nb is not None:
                if nb.address != pwndbg.aglib.regs.pc:
                    # Stop right at the next branch instruction.
                    with proc.break_at(BreakpointLocation(nb.address), internal=True) as bp:
                        await ec.cont(bp)
                else:
                    # Nudge execution so we take the branch we're on top of.
                    pass

        if pwndbg.aglib.proc.alive():
            await ec.single_step()

        # Break on signal as it may be a segfault
        if pwndbg.aglib.proc.stopped_with_signal():
            return False

    return False


async def break_on_program_code(ec: pwndbg.dbg_mod.ExecutionController) -> bool:
    """
    Breaks on next instruction that belongs to process' objfile code

    :return: True for success, False when process ended or when pc is not at the code or if a signal occurred
    """
    exe = pwndbg.aglib.proc.exe()
    binary_exec_page_ranges = tuple(
        (p.start, p.end) for p in pwndbg.aglib.vmmap.get() if p.objfile == exe and p.execute
    )

    pc = pwndbg.aglib.regs.pc
    for start, end in binary_exec_page_ranges:
        if start <= pc < end:
            print(message.error("The pc is already at the binary objfile code. Not stepping."))
            return False

    regs = pwndbg.aglib.regs

    while pwndbg.aglib.proc.alive():
        # Break on signal as it may be a segfault
        if pwndbg.aglib.proc.stopped_with_signal():
            return False

        await break_next_ret(ec)
        await ec.single_step()

        for start, end in binary_exec_page_ranges:
            if start <= regs.pc < end:
                return True

    return False


async def break_on_next(ec: pwndbg.dbg_mod.ExecutionController, address=None) -> None:
    address = address or pwndbg.aglib.regs.pc
    ins = pwndbg.aglib.disasm.disassembly.one(address)

    proc = pwndbg.dbg.selected_inferior()
    with proc.break_at(BreakpointLocation(ins.address + ins.size), internal=True) as bp:
        await ec.cont(bp)



from __future__ import annotations

import argparse

import pwnlib.rop.srop

import pwndbg.aglib
import pwndbg.aglib.memory
import pwndbg.aglib.proc
import pwndbg.chain
import pwndbg.color.context as ctx_color
import pwndbg.color.memory as mem_color
import pwndbg.commands
from pwndbg.commands import CommandCategory
from pwndbg.lib.regs import aarch64
from pwndbg.lib.regs import amd64
from pwndbg.lib.regs import arm
from pwndbg.lib.regs import i386

# Grab frame values from pwntools. Offsets are defined as the offset to stack pointer when syscall instruction is called
# Offsets and names are from Linux kernel source. For example x86_64 is defined in CONFIG_X86_64 struct rt_sigframe (Linux Kernel /arch/x86/include/asm/sigframe.h)
SIGRETURN_FRAME_LAYOUTS: dict[str, list[tuple[int, str]]] = {
    "x86-64": sorted([(-8, "&pretcode")] + list(pwnlib.rop.srop.registers["amd64"].items())),
    "i386": sorted(pwnlib.rop.srop.registers["i386"].items()),
    "aarch64": sorted(pwnlib.rop.srop.registers["aarch64"].items()),
    "arm": sorted(pwnlib.rop.srop.registers["arm"].items()),
}

# Always print these registers (as well as flag register, eflags / cpsr)
SIGRETURN_CORE_REGISTER: dict[str, set[str]] = {
    "x86-64": {*amd64.gpr, amd64.frame, amd64.stack, amd64.pc},
    "i386": {*i386.gpr, i386.frame, i386.stack, i386.pc},
    "aarch64": {*aarch64.gpr, "sp", "pc"},
    "arm": {*arm.gpr, "fp", "ip", "sp", "lr", "pc"},
}


parser = argparse.ArgumentParser(description="Display the SigreturnFrame at the specific address")

parser.add_argument(
    "address", nargs="?", default=None, type=int, help="The address to read the frame from"
)

parser.add_argument(
    "-a",
    "--all",
    dest="display_all",
    action="store_true",
    default=False,
    help="Show all values in the frame in addition to common registers",
)

parser.add_argument(
    "-p",
    "--print",
    dest="print_address",
    action="store_true",
    default=False,
    help="Show addresses of frame values",
)


@pwndbg.commands.Command(parser, category=CommandCategory.MISC)
@pwndbg.commands.OnlyWhenRunning
@pwndbg.aglib.proc.OnlyWithArch(["x86-64", "i386", "aarch64", "arm"])
def sigreturn(address: int = None, display_all=False, print_address=False) -> None:
    address = pwndbg.aglib.regs.sp if address is None else address

    ptr_size = pwndbg.aglib.arch.ptrsize

    frame_layout = SIGRETURN_FRAME_LAYOUTS[pwndbg.aglib.arch.name]
    core_registers = SIGRETURN_CORE_REGISTER[pwndbg.aglib.arch.name]

    # Offset to the stack pointer where the frame values really begins. Start reading memory there.
    # Can be negative, 0, or positive
    frame_start_offset = frame_layout[0][0]

    read_size = frame_layout[-1][0] - frame_start_offset + ptr_size

    mem = pwndbg.aglib.memory.read(address + frame_start_offset, read_size)

    for stack_offset, reg in frame_layout:
        # Subtract the offset of start of frame, to get the correct offset into "mem"
        mem_offset = stack_offset - frame_start_offset

        regname = ctx_color.register(reg.ljust(4).upper())
        value = pwndbg.aglib.arch.unpack(mem[mem_offset : mem_offset + ptr_size])

        if reg in core_registers:
            desc = pwndbg.chain.format(value)

            print_value(f"{regname} {desc}", address + stack_offset, print_address)

        elif reg in pwndbg.aglib.regs.flags:  # eflags or cpsr
            reg_flags = pwndbg.aglib.regs.flags[reg]
            desc = ctx_color.format_flags(value, reg_flags)

            print_value(f"{regname} {desc}", address + stack_offset, print_address)

        elif display_all:
            print_value(f"{reg} {mem_color.get(value)}", address + stack_offset, print_address)


def print_value(string: str, address: int, print_address) -> None:
    addr = ""
    if print_address:
        addr = f"{mem_color.get(address)}: "
    print(f"{addr}{string}")



from __future__ import annotations

import pwndbg
import pwndbg.aglib
import pwndbg.aglib.symbol
import pwndbg.dbg_mod
import pwndbg.libc


def get() -> tuple[int, str]:
    """
    Gets the integer errno value.

    Returns:
        (errno, err_str): If err_str is "", errno is returned as an integer.
        Otherwise, the errno value could not be returned and err_str contains
        the error message.
    """
    # errno is a thread local variable provided by the libc, so we ask the libc where it is.
    # We intentionally avoid expression evaluation because it is not available in corefiles (see #3672).

    # NOTE: Maybe this should be moved to pwndbg.libc

    # The definition of errno in glibc is like this:
    # # define errno (*__errno_location ())
    # https://elixir.bootlin.com/glibc/glibc-2.41/source/stdlib/errno.h#L38
    # int *
    # __errno_location (void)
    # {
    #     return &errno;
    # }
    # https://elixir.bootlin.com/glibc/glibc-2.41/source/csu/errno-loc.c#L24
    # #   define errno __libc_errno
    # https://elixir.bootlin.com/glibc/glibc-2.41/source/include/errno.h#L27
    # extern __thread int __libc_errno __attribute__ ((alias ("errno")))
    # https://elixir.bootlin.com/glibc/glibc-2.41/source/csu/errno.c#L32
    # In glibc, errno is always available as a thread local variable (it is an exported/dynamic
    # symbol).

    # The definition of errno in musl is like this:
    # #define errno (*__errno_location())
    # https://elixir.bootlin.com/musl/v1.2.5/source/include/errno.h#L16
    # int *__errno_location(void)
    # {
    # 	return &__pthread_self()->errno_val;
    # }
    # https://elixir.bootlin.com/musl/v1.2.5/source/src/errno/__errno_location.c#L4
    # In musl, there is no errno symbol, but __errno_location() is an exported/dynamic
    # function symbol.

    err_str: str = ""
    try:
        maybe_errno: int | None = pwndbg.aglib.symbol.lookup_symbol_value(
            "errno",
            type=pwndbg.dbg_mod.SymbolLookupType.VARIABLE,
            objfile_endswith=str(pwndbg.libc.filepath()),
        )
        if maybe_errno is not None:
            return maybe_errno, ""

        err_str = "TLS variable `errno` not found in the libc symbol table."
    except pwndbg.dbg_mod.Error as e:
        err_str = str(e)

    # We don't really need to do this, but we do it for a better diagnostic.
    errno_location: pwndbg.dbg_mod.Value | None = pwndbg.aglib.symbol.lookup_symbol(
        "__errno_location",
        type=pwndbg.dbg_mod.SymbolLookupType.FUNCTION,
        objfile_endswith=str(pwndbg.libc.filepath()),
    )
    if errno_location is None:
        err_str += "\nFunction `__errno_location` not found in the libc mapping."
        return -1, err_str

    # If we are a corefile we cannot execute expressions.
    if pwndbg.dbg.selected_inferior().is_core_file():
        err_str += (
            "\nFunction `__errno_location` found, but we cannot execute it as the target is a\n"
        )
        err_str += "corefile. You may look at the assembly `x/10i __errno_location` and figure out the TLS\n"
        err_str += "offset manually. Then `x/wx $fs_base + offset`."
        return -1, err_str

    # Try executing __errno_location().
    frame = pwndbg.dbg.selected_frame()
    try:
        if frame is not None:
            return int(
                frame.evaluate_expression(
                    "*((int *(*) (void)) __errno_location) ()", lock_scheduler=True
                )
            ), ""
        return int(
            pwndbg.dbg.selected_inferior().evaluate_expression(
                "*((int *(*) (void)) __errno_location) ()"
            )
        ), ""
    except pwndbg.dbg_mod.Error as e:
        err_str += f"\nFailed executing __errno_location(): {e}"
        return -1, err_str



from __future__ import annotations

from .dispatch import LibcProvider
from .dispatch import LibcType
from .dispatch import LibcURLs
from .facade import addr
from .facade import filepath
from .facade import has_debug_info
from .facade import has_exported_symbols
from .facade import has_internal_symbols
from .facade import loader_addr
from .facade import loader_filepath
from .facade import relocations_by_section_name
from .facade import section_address_by_name
from .facade import section_by_name
from .facade import urls
from .facade import version
from .facade import which

__all__ = [
    "LibcProvider",
    "LibcType",
    "LibcURLs",
    "which",
    "addr",
    "filepath",
    "has_debug_info",
    "has_exported_symbols",
    "has_internal_symbols",
    "loader_addr",
    "loader_filepath",
    "relocations_by_section_name",
    "section_address_by_name",
    "section_by_name",
    "urls",
    "version",
]



"""
Radare2 integration with r2pipe.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import pwndbg.aglib.elf
import pwndbg.aglib.kernel
import pwndbg.aglib.proc
import pwndbg.aglib.qemu
import pwndbg.lib.cache
from pwndbg.color import message
from pwndbg.lib import ErrorCode
from pwndbg.lib import Status

if TYPE_CHECKING:
    import r2pipe as r2pipe_mod


@pwndbg.lib.cache.cache_until("start", "objfile")
def r2pipe() -> r2pipe_mod.open | Status:
    """
    Spawn and return a r2pipe handle for the current process file.

    This function requires a radare2 installation plus the r2pipe python
    library. The base address is automatically set for PIE when loading the
    binary.
    After opening the handle, the binary is automatically analyzed.

    If successful, returns the handle, otherwise returns a failing Status.
    """
    filename = pwndbg.aglib.proc.exe()
    if filename is None:
        return Status.fail("Could not find objfile to create a r2pipe for.")

    try:
        import r2pipe as r2pipe_mod
    except ImportError:
        return Status.coded_fail(
            ErrorCode.NO_IMPORT, "Could not import r2pipe python library. Is it installed?"
        )

    if pwndbg.aglib.qemu.is_qemu_kernel():
        flags = ["-e", "bin.cache=true", "-e", "bin.relocs.apply=true"]
        if (kbase := pwndbg.aglib.kernel.kbase()) and filename == pwndbg.aglib.proc.exe():
            urand = pwndbg.aglib.elf.get_vmlinux_unrand_base(filename)
            if urand is not None:
                flags.extend(
                    [
                        "-e",
                        "bin.baddr=" + hex(kbase - urand),
                    ]
                )
        r2 = r2pipe_mod.open(filename, flags)
    else:
        flags = ["-e", "io.cache=true"]
        exe = pwndbg.aglib.elf.exe()
        if pwndbg.aglib.elf.get_elf_info(filename).is_pie and exe is not None:
            flags.extend(["-B", hex(exe.address)])
        r2 = r2pipe_mod.open(filename, flags=flags)
    # LD -> list supported decompilers (e cmd.pdc=?)
    # Outputs for example: pdc\npdg
    if "pdg" not in str(r2.cmd("LD")).split("\n"):
        return Status.coded_fail(
            ErrorCode.NO_IMPORT, "radare2 plugin r2ghidra must be installed and available from r2."
        )
    return r2


def r2cmd(arguments: list[str]) -> str:
    """
    Return result of rizin command or error string.
    """
    r2 = r2pipe()
    if isinstance(r2, Status):
        # Since we got a status, it must be a failure.
        return message.error(r2.message)
    return str(r2.cmd(" ".join(arguments)))



from __future__ import annotations

import argparse
import math
import re

from capstone6pwndbg import CS_ARCH_BPF
from capstone6pwndbg import CS_MODE_BPF_EXTENDED
from capstone6pwndbg import CS_MODE_LITTLE_ENDIAN
from capstone6pwndbg import Cs

import pwndbg
import pwndbg.aglib.kernel
import pwndbg.aglib.kernel.bpf
import pwndbg.aglib.memory
import pwndbg.aglib.typeinfo
import pwndbg.color.message as message
import pwndbg.commands
from pwndbg.commands import CommandCategory
from pwndbg.lib.exception import IndentContextManager

parser = argparse.ArgumentParser(
    description="Prints information about the linux kernel bpf progs and maps."
)
parser.add_argument("-v", "--verbose", action="count", default=0, help="Increase the verbosity.")
parser.add_argument(
    "-p", "--progs", dest="print_progs", action="store_true", default=False, help="Print progs"
)
parser.add_argument(
    "-m", "--maps", dest="print_maps", action="store_true", default=False, help="Print maps"
)

_bpf_map_array_off = None
MAX_PRINTED_VALUE_SIZE = 0x20
MAX_BPF_VERBOSE_LEVEL1_OUTPUT_LEN = 0x10
BPF_FIRST_REG, BPF_SECOND_REG = 1 << 0, 1 << 1
BPF_AUX_REG_STRING = "ax"
BPF_MAP_ARRAY_TYPES = (
    "BPF_MAP_TYPE_ARRAY",
    "BPF_MAP_TYPE_PROG_ARRAY",
)


def handle_bpf_aux_reg_for_insns_bytes(insns_bytes):
    # https://elixir.bootlin.com/linux/v6.17.1/source/include/linux/filter.h#L62
    sz = len(insns_bytes)
    result = [0] * (len(insns_bytes) // 8)
    for i in range(1, sz, 8):
        b = insns_bytes[i]
        if b & 0xF == 0xB:
            result[i // 8] |= BPF_FIRST_REG
            insns_bytes[i] &= ~0xF
        if b & 0xF0 == 0xB0:
            result[i // 8] |= BPF_SECOND_REG
            insns_bytes[i] &= ~0xF0
    return result


def handle_bpf_aux_reg_for_opstr(opstr, regflag):
    if regflag == 0:
        return opstr
    pattern = re.compile(r"r0")
    matches = list(pattern.finditer(opstr))
    if regflag & BPF_FIRST_REG:
        start, end = matches[0].span()
        opstr = opstr[:start] + BPF_AUX_REG_STRING + opstr[end:]
    if regflag & BPF_SECOND_REG:
        start, end = matches[-1].span()
        opstr = opstr[:start] + BPF_AUX_REG_STRING + opstr[end:]
    return opstr


def bpf_map_array_offset(bpf_array, t, max_entries, value_size):
    global _bpf_map_array_off
    if _bpf_map_array_off:
        # pwndbg.lib.cache is not used here because it would also cache None
        return _bpf_map_array_off
    if t in BPF_MAP_ARRAY_TYPES:
        expected_elem_size = math.ceil(value_size / 8) * 8
        expected_index_mask = (1 << math.ceil(math.log2(max_entries))) - 1
        for i in range(200):
            elem_size = pwndbg.aglib.memory.u32(bpf_array + 4 * i)
            index_mask = pwndbg.aglib.memory.u32(bpf_array + 4 * (i + 1))
            if elem_size == expected_elem_size and index_mask == expected_index_mask:
                """
                struct bpf_array {
                    struct bpf_map map;
                    u32 elem_size; // i points here
                    u32 index_mask;
                    struct bpf_array_aux *aux;
                    union {
                        DECLARE_FLEX_ARRAY(char, value) __aligned(8);
                        DECLARE_FLEX_ARRAY(void *, ptrs) __aligned(8);
                        DECLARE_FLEX_ARRAY(void __percpu *, pptrs) __aligned(8);
                    };
                };
                """
                _bpf_map_array_off = (i + 2) * 4 + pwndbg.aglib.arch.ptrsize
                break
    return _bpf_map_array_off


def parse_xa_node(xa_node):
    xa_node = int(xa_node) & ~3
    if not pwndbg.aglib.memory.is_kernel(xa_node):
        return []
    xa_node = pwndbg.aglib.memory.get_typed_pointer("struct xa_node", xa_node)
    result = []
    shift = int(xa_node["shift"])
    count = int(xa_node["count"])
    for i in range(64):
        slot = int(xa_node["slots"][i])
        if slot == 0:
            continue
        if shift:
            result += parse_xa_node(slot)
        else:
            result.append(slot)
        count -= 1
        if count == 0:
            break
    return result


def print_bpf_progs(verbose):
    indent = IndentContextManager()
    prog_idr = pwndbg.aglib.kernel.prog_idr()
    if int(prog_idr) == 0:
        print(message.warn("cannot find prog_idr"))
        return
    prog_idr = pwndbg.aglib.memory.get_typed_pointer("struct idr", prog_idr)
    xa_node = prog_idr["idr_rt"]["xa_head"]
    indent.print(indent.prefix("bpf progs") + f": prog_idr @ {indent.addr_hex(int(prog_idr))}")
    if int(xa_node) == 0:
        return
    slots = parse_xa_node(xa_node)
    with indent:
        for idx, slot in enumerate(slots):
            bpf_prog = pwndbg.aglib.memory.get_typed_pointer("struct bpf_prog", slot)
            t = bpf_prog["type"].value_to_human_readable()
            attach_t = bpf_prog["expected_attach_type"].value_to_human_readable()
            prefix = indent.prefix(f"[0x{idx:02x}] {indent.addr_hex(slot)}")
            indent.print(
                f"{prefix} (type: {message.success(t)}, attach: {message.success(attach_t)})"
            )
            with indent:
                func = int(bpf_prog["bpf_func"])
                aux = int(bpf_prog["aux"])
                jited_len = int(bpf_prog["jited_len"])
                desc = f"func @ {indent.aux_hex(func)} (jited_len: {indent.aux_hex(jited_len)}), aux @ {indent.aux_hex(aux)}"
                indent.print(desc)
                if verbose > 0:
                    cs = Cs(
                        CS_ARCH_BPF,
                        CS_MODE_LITTLE_ENDIAN | CS_MODE_BPF_EXTENDED,
                    )
                    num_insns = int(bpf_prog["len"])
                    insns = int(bpf_prog["insns"].address)
                    insns_bytes = pwndbg.aglib.memory.read(insns, num_insns * 8)
                    aux_regs = handle_bpf_aux_reg_for_insns_bytes(insns_bytes)
                    with indent:
                        indent.print(indent.prefix(f"{num_insns} insns") + ":")
                        for i in range(num_insns):
                            if i == MAX_BPF_VERBOSE_LEVEL1_OUTPUT_LEN and verbose == 1:
                                indent.print("... (truncated)")
                                indent.print(
                                    message.warn("max output len reached, use -vv for full output")
                                )
                                break
                            off = i * 8
                            address = insns + off
                            disass = list(
                                cs.disasm(bytes(insns_bytes[off : off + 8]), insns + address)
                            )
                            if len(disass) == 0:
                                bytecode = ""
                                for b in insns_bytes[off : off + 8]:
                                    bytecode += f"{b:02x} "
                                desc = message.error(f"invalid insn: {bytecode}")
                                indent.print(f"{indent.addr_hex(address)}\t{desc}")
                                continue
                            insn = disass[0]
                            mnemonic = insn.mnemonic
                            opstr = insn.op_str
                            opstr = handle_bpf_aux_reg_for_opstr(opstr, aux_regs[i])
                            indent.print(f"{indent.addr_hex(address)}\t{mnemonic}\t{opstr}")


def print_bpf_maps(verbose):
    indent = IndentContextManager()
    map_idr = pwndbg.aglib.kernel.map_idr()
    if int(map_idr) == 0:
        print(message.warn("cannot find map_idr"))
        return
    map_idr = pwndbg.aglib.memory.get_typed_pointer("struct idr", map_idr)
    xa_node = map_idr["idr_rt"]["xa_head"]
    if int(xa_node) == 0:
        return
    indent.print(indent.prefix("bpf maps") + f": map_idr @ {indent.addr_hex(int(map_idr))}")
    slots = parse_xa_node(xa_node)
    with indent:
        for idx, slot in enumerate(slots):
            bpf_array = pwndbg.aglib.memory.get_typed_pointer("struct bpf_array", slot)
            prefix = indent.prefix(f"[0x{idx:02x}] {indent.addr_hex(slot)}")
            t = bpf_array["map"]["map_type"].value_to_human_readable()
            indent.print(f"{prefix} (type: {message.success(t)})")
            with indent:
                key_size = int(bpf_array["map"]["key_size"])
                value_size = int(bpf_array["map"]["value_size"])
                max_entries = int(bpf_array["map"]["max_entries"])
                bpf_array = int(bpf_array)
                off = bpf_map_array_offset(bpf_array, t, max_entries, value_size)
                content = indent.aux_hex(bpf_array + off) if off else "unknown"
                desc = f"array @ {content} (key_size: {indent.aux_hex(key_size)}, value_size: {indent.aux_hex(value_size)}, max_entries: {indent.aux_hex(max_entries)})"
                indent.print(desc)
                # TODO: what about types other than array
                if off is not None and verbose > 0 and t in BPF_MAP_ARRAY_TYPES:
                    with indent:
                        entrysz = math.ceil(value_size / 8) * 8
                        for i in range(max_entries):
                            if i == MAX_BPF_VERBOSE_LEVEL1_OUTPUT_LEN and verbose == 1:
                                indent.print("... (truncated)")
                                indent.print(
                                    message.warn("max output len reached, use -vv for full output")
                                )
                                break
                            idxfmt = f"[0x{i:02x}]"
                            sz = min(value_size, MAX_PRINTED_VALUE_SIZE)
                            value = ""
                            for b in pwndbg.aglib.memory.read(bpf_array + off + i * entrysz, sz):
                                value += f"{b:02x} "
                            if sz < value_size:
                                value += "... (" + message.warn("truncated") + ")"
                            indent.print(f"- {indent.prefix(idxfmt)} {value}")


@pwndbg.commands.Command(parser, category=CommandCategory.KERNEL)
@pwndbg.commands.OnlyWhenQemuKernel
@pwndbg.commands.OnlyWithKernelSymbols
@pwndbg.commands.OnlyWhenPagingEnabled
@pwndbg.commands.WarnOnKernelConfigRandstruct
def kbpf(verbose: int, print_progs: bool, print_maps: bool) -> None:
    pwndbg.aglib.kernel.bpf.recover_bpf_typeinfo()
    if pwndbg.aglib.typeinfo.load("struct idr") is None:
        return
    if not print_progs and not print_maps:
        print_progs = print_maps = True
    if print_progs:
        print_bpf_progs(verbose)
    if print_maps:
        print_bpf_maps(verbose)



"""
Function and syscall argument formatting.

Enumerates arguments which may be passed in a combination of
registers and stack values.
"""

from __future__ import annotations

import re

from capstone6pwndbg import CS_GRP_INT

import pwndbg.aglib
import pwndbg.aglib.file
import pwndbg.aglib.memory
import pwndbg.aglib.objc
import pwndbg.aglib.proc
import pwndbg.aglib.symbol
import pwndbg.chain
import pwndbg.dbg_mod
import pwndbg.enhance
import pwndbg.lib.abi
import pwndbg.lib.functions
from pwndbg.aglib.disasm.instruction import PwndbgInstruction
from pwndbg.aglib.nearpc import c as N
from pwndbg.lib.arch import Platform
from pwndbg.lib.functions import Function
from pwndbg.lib.functions import format_flags_argument


def get(instruction: PwndbgInstruction) -> list[tuple[pwndbg.lib.functions.Argument, int]]:
    """
    Returns an array containing the arguments to the current function,
    if $pc is a function call or syscall instruction.

    Otherwise, returns None.
    """

    if instruction is None:
        return []

    if instruction.address != pwndbg.aglib.regs.pc:
        return []

    if instruction.call_like:
        abi = pwndbg.aglib.arch.function_abi

        if abi is None:
            return []

        target = instruction.target

        if not target:
            return []

        name = pwndbg.aglib.symbol.resolve_addr(target)
        if not name:
            return []
    elif CS_GRP_INT in instruction.groups:
        # Get the syscall number and name
        name = instruction.syscall_name
        abi = pwndbg.aglib.arch.syscall_abi
        target = None

        if name is None or abi is None:
            return []
    else:
        return []

    original_name = name or ""

    name = original_name.replace("isoc99_", "")  # __isoc99_sscanf
    name = name.replace("@plt", "")  # getpwiod@plt

    # If we have particular `XXX_chk` function in our database, we use it.
    # Otherwise, we show args for its unchecked version.
    # We also lstrip `_` in here, as e.g. `__printf_chk` needs the underscores.
    if name not in pwndbg.lib.functions.functions:
        name = name.replace("_chk", "")
        name = name.strip().lstrip("_")  # _malloc

    func: Function | None = None
    if pwndbg.aglib.arch.platform == Platform.DARWIN:
        # Try to resolve an Objective-C method call.
        #
        # Checking this first keeps us from resolving these as simple calls to
        # `objc_msgSend` and functions like it, which have definitions that are
        # rather barren of semantics in comparison.
        func = pwndbg.aglib.objc.try_resolve_call_at_current_pc(instruction)

    if func is None:
        # If more specific call information can't be determined, use the regular
        # function resolution flow.
        func = pwndbg.lib.functions.functions.get(name, None)

    # FIXME(provider, integration): Add this feature back at some point
    # Try to grab the data out of IDA
    # if not func and target:
    #    func = pwndbg.dintegration.provider.get_func_type(target)

    if func:
        args = func.args
        if len(args) > 1 and args[-1].name == "vararg":
            format_value = pwndbg.enhance.enhance(argument(len(args) - 2, abi))
            m = re.findall(
                r"%[-+ #0]?(?:[0-9]+|\*)?(?:\.(?:[0-9]+|\*))?(?:hh|h|l|ll|q|L|j|z|Z|t)?[diuoxXfFeEgGaAcsCSpn]",
                format_value,
            )
            vararg_cnt = len(m)
            if vararg_cnt > 0:
                args.pop()
                args += [
                    pwndbg.lib.functions.Argument("int", 0, argname(len(args) + i, abi))
                    for i in range(vararg_cnt)
                ]
    else:
        n_args_default = 4
        sym = pwndbg.aglib.symbol.lookup_frame_symbol(original_name)
        if sym:
            try:
                target_type = sym.type.target()
            except Exception:
                target_type = sym.type

            if target_type and target_type.code == pwndbg.dbg_mod.TypeCode.FUNC:
                func_args = target_type.func_arguments()
                if func_args is not None:
                    n_args_default = len(func_args)
        args = (
            pwndbg.lib.functions.Argument("int", 0, argname(i, abi)) for i in range(n_args_default)
        )

    return [(arg, argument(i, abi)) for i, arg in enumerate(args)]


def argname(n: int, abi: pwndbg.lib.abi.ABI) -> str:
    regs = abi.register_arguments

    if n < len(regs):
        return regs[n]

    return f"arg[{n}]"


def argument(n: int, abi: pwndbg.lib.abi.ABI | None = None) -> int:
    """
    Returns the nth argument, as if $pc were a 'call' or 'bl' type
    instruction.
    Works only for ABIs that use registers for arguments.
    """
    abi = abi or pwndbg.aglib.arch.function_abi
    if abi is None:
        raise pwndbg.dbg_mod.Error(
            f"Function ABI not defined for current architecture, {pwndbg.aglib.arch.function_abi}"
        )
    regs = abi.register_arguments

    if n < len(regs):
        return pwndbg.aglib.regs.read_reg_uncached(regs[n])

    n -= len(regs)

    sp = pwndbg.aglib.regs.sp + (n * pwndbg.aglib.arch.ptrsize)

    return pwndbg.aglib.memory.read_pointer_width(sp)


def arguments(abi: pwndbg.lib.abi.ABI | None = None):
    """
    Yields (arg_name, arg_value) tuples for arguments from a given ABI.
    Works only for ABIs that use registers for arguments.
    """
    abi = abi or pwndbg.aglib.arch.function_abi
    if abi is None:
        return []
    regs = abi.register_arguments

    for i in range(len(regs)):
        yield argname(i, abi), argument(i, abi)


# When an argument is named one of these in Linux syscalls/glibc, it refers to a file descriptor
# Search for strings containing "fd" in https://chromium.googlesource.com/chromiumos/docs/+/master/constants/syscalls.md
FILE_DESCRIPTOR_ARG_NAMES = {
    "fd",
    "in_fd",
    "out_fd",
    "fdin",
    "fdout",
    "oldfd",
    "fildes",
    "newfd",
    "epfd",
    "dfd",
    "dirfd",
    "mountdirfd",
}


def format_args(instruction: PwndbgInstruction) -> list[str]:
    result = []
    for arg, value in get(instruction):
        code = arg.type != "char"
        pretty = (
            pwndbg.chain.format(value, code=code)
            if not arg.flags
            else format_flags_argument(arg.flags, value)
        )

        # Enhance args display
        if arg.name in FILE_DESCRIPTOR_ARG_NAMES and isinstance(value, int):
            # Cannot find PID of the QEMU program: perhaps it is in a different pid namespace or we have no permission to read the QEMU process' /proc/$pid/fd/$fd file.
            pid = pwndbg.aglib.proc.pid()
            if pid is not None:
                path = pwndbg.aglib.file.readlink(f"/proc/{pid}/fd/{value}")
                if path:
                    pretty += f" ({path})"

        result.append(f"{N.argument(arg.name) + ':':<10} {pretty}")

    return result
